#!/usr/bin/env python3
"""RAK811 P2P demo.

Send counter messages at random interval and listen the rest of the time.

Start this script on 2 or more nodes an observe the packets flowing.

Copyright 2019, 2021 Philippe Vanhaesendonck

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

SPDX-License-Identifier: Apache-2.0
"""
from random import randint
from sys import exit
from time import time

from rak811.rak811 import Mode, Rak811

# Send packet every P2P_BASE + (0..P2P_RANDOM) seconds
P2P_BASE = 30
P2P_RANDOM = 60

# Magic key to recognize our messages
P2P_MAGIC = b'\xca\xfe'

lora = Rak811()

# Most of the setup should happen only once...
print('Setup')
lora.hard_reset()
lora.mode = Mode.LoRaP2P

# RF configuration
# - Avoid LoRaWan channels (You will get quite a lot of spurious packets!)
# - Respect local regulation (frequency, power, duty cycle)
lora.rf_config = {
    'sf': 7,
    'freq': 869.800,
    'pwr': 16
}

print('Entering send/receive loop')
counter = 0
try:
    while True:
        # Calculate next message send timestamp
        next_send = time() + P2P_BASE + randint(0, P2P_RANDOM)
        # Set module in receive mode
        lora.rxc()
        # Loop until we reach the next send time
        # Don't enter loop for small wait times (<1 1 sec.)
        while (time() + 1) < next_send:
            wait_time = next_send - time()
            print('Waiting on message for {:0.0f} seconds'.format(wait_time))
            # Note that you don't have to listen actively for capturing message
            # Once in receive mode, the library will capture all messages sent.
            lora.rx_get(wait_time)
            while lora.nb_downlinks > 0:
                message = lora.get_downlink()
                data = message['data']
                if data[:len(P2P_MAGIC)] == P2P_MAGIC:
                    print(
                        'Received message: {}'.format(
                            int.from_bytes(data[len(P2P_MAGIC):],
                                           byteorder='big')
                        )
                    )
                    print('RSSI: {}, SNR: {}'.format(message['rssi'],
                                                     message['snr']))
                else:
                    print('Foreign message received')
        # Time to send message
        # Exit receive mode
        lora.rx_stop()
        counter += 1
        print('Send message {}'.format(counter))
        lora.txc(P2P_MAGIC + bytes.fromhex('{:08x}'.format(counter)))

except:  # noqa: E722
    pass

print('All done')
exit(0)



"""Units tests for the Rak811Serial class."""
from time import sleep

from mock import Mock, patch
from pytest import raises
from serial import EIGHTBITS
# Ignore RPi.GPIO
p = patch.dict('sys.modules', {'RPi': Mock()})
p.start()
from rak811.serial import BAUDRATE, PORT, TIMEOUT  # noqa: E402, I100
from rak811.serial import Rak811Serial, Rak811TimeoutError  # noqa: E402


@patch('rak811.serial.Serial')
def test_instantiate_default(mock_serial):
    """Test that Rak811Serial can be instantiated.

    Check for basic initialisation and teardown of the serial interface.
    """
    mock_serial.return_value.readline.return_value = b''
    rs = Rak811Serial()
    # Test default parameters are used
    mock_serial.assert_called_once_with(port=PORT,
                                        baudrate=BAUDRATE,
                                        timeout=TIMEOUT)
    # Class initialization
    mock_serial.return_value.reset_input_buffer.assert_called_once()
    assert rs._alive

    # Test tear down
    rs.close()
    mock_serial.return_value.close.assert_called_once()
    assert not rs._alive


@patch('rak811.serial.Serial')
def test_instantiate_custom(mock_serial):
    """Test that Rak811Serial can be instantiated - custom parameters."""
    mock_serial.return_value.readline.return_value = b''
    port = '/dev/ttyAMA0'
    timeout = 5
    bytesize = EIGHTBITS
    rs = Rak811Serial(port=port, timeout=timeout, bytesize=bytesize)

    mock_serial.assert_called_once_with(
        port=port,
        baudrate=BAUDRATE,
        timeout=timeout,
        bytesize=bytesize
    )
    rs.close()


@patch('rak811.serial.Serial')
def test_send_string(mock_serial):
    """Test Rak811Serial.send_string."""
    mock_serial.return_value.readline.return_value = b''
    rs = Rak811Serial()

    rs.send_string('Hello world')
    mock_serial.return_value.write.assert_called_once_with(b'Hello world')

    rs.close()


@patch('rak811.serial.Serial')
def test_send_command(mock_serial):
    """Test Rak811Serial.send_command."""
    mock_serial.return_value.readline.return_value = b''
    rs = Rak811Serial()

    rs.send_command('RESET')
    mock_serial.return_value.write.assert_called_once_with(b'at+RESET\r\n')

    rs.close()


def emulate_rak_input(mock, timeout, data_in):
    """Emulate Rak811 output.

    Parameters:
        mock: mocked Serial class
        timeout is the Serial.readline() timeout
        data_in is a list of tuples (delay, bytes):
            delay: delay before data is available
            bytes: output from the Rak811 module

    When used, this function needs to be called before instantiating the
    Rak811Serial object, as it starts the read thread immediately.

    """
    def side_effect():
        if len(data):
            (d1, b1) = data.pop(0)
            sleep(d1)
            # If we have data readily available after this one, set in_waiting
            # to the data size.
            # If we are at the end, or if there is a delay before next set
            # in_waiting to zero.
            if len(data):
                (d2, b2) = data[0]
                if d2:
                    type(mock.return_value).in_waiting = 0
                else:
                    type(mock.return_value).in_waiting = len(b2)
            else:
                type(mock.return_value).in_waiting = 0
            return b1
        else:
            # No more data
            sleep(timeout)
            return b''

    # Take a local copy of the list and update mock
    data = list(data_in)
    mock.return_value.readline.side_effect = side_effect
    # Set in_waiting to the size of the first data line
    if len(data):
        type(mock.return_value).in_waiting = len(data[0][1])
    else:
        type(mock.return_value).in_waiting = 0


@patch('rak811.serial.Serial')
def test_receive_single(mock_serial):
    """Test Rak811Serial.receive, single line."""
    # For response, first line is passed, others, if any are buffered.
    emulate_rak_input(mock_serial, 1, [
        (0, b'OK\r\n'),
        (0, b'OKok\r\n'),
    ])
    rs = Rak811Serial()
    assert rs.receive() == 'OK'
    assert rs.receive() == 'OKok'
    rs.close()

    # Check for Errors
    emulate_rak_input(mock_serial, 1, [
        (0, b'ERROR-1\r\n'),
    ])
    rs = Rak811Serial()
    assert rs.receive() == 'ERROR-1'
    rs.close()

    # Noise is skipped
    emulate_rak_input(mock_serial, 1, [
        (0, b'Welcome to RAK811\r\n'),
        (0.5, b'\r\n'),
        (0, b'\r\n'),
        (0, b'OK\r\n'),
    ])
    rs = Rak811Serial()
    assert rs.receive() == 'OK'
    rs.close()

    # Noise is returned
    emulate_rak_input(mock_serial, 1, [
        (0, b'Welcome to RAK811\r\n'),
        (0, b'OK\r\n'),
    ])
    rs = Rak811Serial(keep_untagged=True)
    assert rs.receive() == 'Welcome to RAK811'
    assert rs.receive() == 'OK'
    rs.close()

    # Handle non-ASCII characters
    emulate_rak_input(mock_serial, 1, [
        (0, b'Non ASCII: \xde\xad\xbe\xef\r\n'),
        (0.5, b'\r\n'),
        (0, b'\r\n'),
        (0, b'OK\r\n'),
    ])
    rs = Rak811Serial()
    assert rs.receive() == 'OK'
    rs.close()

    # Response timeout
    emulate_rak_input(mock_serial, 1, [
    ])
    rs = Rak811Serial(read_buffer_timeout=1)
    with raises(Rak811TimeoutError,
                match='Timeout while waiting for data'):
        rs.receive()
    rs.close()


@patch('rak811.serial.Serial')
def test_receive_multi(mock_serial):
    """Test Rak811Serial.receive, multiple lines."""
    # Single command
    emulate_rak_input(mock_serial, 1, [
        (0, b'at+recv=8,0,0\r\n'),
    ])
    rs = Rak811Serial()
    event = rs.receive(single=False)
    assert len(event) == 1
    assert event.pop() == 'at+recv=8,0,0'
    rs.close()

    # Multiple commands
    emulate_rak_input(mock_serial, 1, [
        (0, b'at+recv=2,0,0\r\n'),
        (0, b'Welcome to RAK811\r\n'),
        (0, b'at+recv=0,0,0\r\n'),
    ])
    rs = Rak811Serial()
    event = rs.receive(single=False)
    assert len(event) == 2
    assert event.pop() == 'at+recv=0,0,0'
    assert event.pop() == 'at+recv=2,0,0'
    rs.close()

    # Event timeout
    emulate_rak_input(mock_serial, 1, [
    ])
    rs = Rak811Serial(read_buffer_timeout=1)
    with raises(Rak811TimeoutError,
                match='Timeout while waiting for data'):
        rs.receive(single=False)
    rs.close()


@patch('rak811.serial.Serial')
def test_get_response_event(mock_serial):
    """Test response / event sequence."""
    # All at once
    emulate_rak_input(mock_serial, 1, [
        (0, b'Welcome 1\r\n'),
        (0, b'OK\r\n'),
        (0, b'Welcome 2\r\n'),
        (0, b'at+recv=2,0,0\r\n'),
        (0, b'Welcome 3\r\n'),
        (0, b'at+recv=0,0,0\r\n'),
    ])
    rs = Rak811Serial()
    assert rs.receive(single=True) == 'OK'
    event = rs.receive(single=False)
    assert len(event) == 2
    assert event.pop() == 'at+recv=0,0,0'
    assert event.pop() == 'at+recv=2,0,0'
    rs.close()

    # Same scenario with delay between response and event
    emulate_rak_input(mock_serial, 1, [
        (0, b'Welcome 1\r\n'),
        (0, b'OK\r\n'),
        (0, b'Welcome 2\r\n'),
        (1, b'at+recv=2,0,0\r\n'),
        (0, b'Welcome 3\r\n'),
        (0, b'at+recv=0,0,0\r\n'),
    ])
    rs = Rak811Serial()
    assert rs.receive(single=True) == 'OK'
    event = rs.receive(single=False)
    assert len(event) == 2
    assert event.pop() == 'at+recv=0,0,0'
    assert event.pop() == 'at+recv=2,0,0'
    rs.close()



"""RAK811 CLI interface.

Provides a command line interface for the RAK811 module (Firmware V2.0).

Copyright 2019, 2021 Philippe Vanhaesendonck

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

SPDX-License-Identifier: Apache-2.0
"""
from json import dumps
import logging

import click

from .rak811 import Mode, RecvEx, Reset
from .rak811 import Rak811
from .rak811 import Rak811Error
from .rak811 import Rak811EventError, Rak811ResponseError, Rak811TimeoutError

# Valid configuration keys for LoRaWan
LW_CONFIG_KEYS = ('dev_addr', 'dev_eui', 'app_eui', 'app_key', 'nwks_key',
                  'apps_key', 'tx_power', 'pwr_level', 'adr', 'dr',
                  'public_net', 'rx_delay1', 'ch_list', 'ch_mask', 'max_chs',
                  'rx2', 'join_cnt', 'nbtrans', 'retrans', 'class', 'duty')

# Valid configuration keys for LoRaP2P
P2P_CONFIG_KEYS = {
    'freq': click.FloatRange(min=860.000, max=929.900, clamp=False),
    'sf': click.IntRange(min=6, max=12, clamp=False),
    'bw': click.IntRange(min=0, max=2, clamp=False),
    'cr': click.IntRange(min=1, max=4, clamp=False),
    'prlen': click.IntRange(min=8, max=65535, clamp=False),
    'pwr': click.IntRange(min=5, max=20, clamp=False)
}


class KeyValueParamTypeLW(click.ParamType):
    """Basic KEY=VALUE pair parameter type for LoRaWan."""

    name = 'key-value-lorawan'

    def convert(self, value, param, ctx):
        try:
            (k, v) = value.split('=')
            k = k.lower()
            if k not in LW_CONFIG_KEYS:
                self.fail('{0} is not a valid config key'.format(k),
                          param,
                          ctx)
            return (k, v)
        except ValueError:
            self.fail('{0} is not a valid Key=Value parameter'.format(value),
                      param,
                      ctx)


class KeyValueParamTypeP2P(click.ParamType):
    """Basic KEY=VALUE pair parameter type for LoRaP2P."""

    name = 'key-value-p2p'

    def convert(self, value, param, ctx):
        try:
            (k, v) = value.split('=')
            k = k.lower()
        except ValueError:
            self.fail('{0} is not a valid Key=Value parameter'.format(value),
                      param,
                      ctx)
        if k not in P2P_CONFIG_KEYS:
            self.fail('{0} is not a valid config key'.format(k),
                      param,
                      ctx)
        v = P2P_CONFIG_KEYS[k].convert(v, param, ctx)
        return (k, v)


def print_exception(e):
    """Print exception raised by the Rak811 library."""
    if isinstance(e, Rak811ResponseError):
        click.echo('RAK811 response error {}: {}'.format(e.errno, e.strerror))
    elif isinstance(e, Rak811EventError):
        click.echo('RAK811 event error {}: {}'.format(e.errno, e.strerror))
    elif isinstance(e, Rak811TimeoutError):
        click.echo('RAK811 timeout: {}'.format(e))
    else:
        click.echo('RAK811 unexpected exception {}'.format(e))


@click.group()
@click.option(
    '-v',
    '--verbose',
    is_flag=True,
    help='Verbose mode'
)
@click.option(
    '-d',
    '--debug',
    is_flag=True,
    help='Debug mode'
)
@click.version_option()
@click.pass_context
def cli(ctx, verbose, debug):
    """Command line interface for the RAK811 module."""
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)


@cli.command(name='hard-reset')
@click.pass_context
def hard_reset(ctx):
    """Hardware reset of the module.

    Hard reset should not be required in normal operation. It needs to be
    issued once after host boot, or module restart.
    """
    lora = Rak811()
    lora.hard_reset()
    if ctx.obj['VERBOSE']:
        click.echo('Hard reset complete')
    lora.close()


"""System commands."""


@cli.command()
@click.pass_context
def version(ctx):
    """Get module version."""
    lora = Rak811()
    click.echo(lora.version)
    lora.close()


@cli.command()
@click.pass_context
def sleep(ctx):
    """Enter sleep mode."""
    lora = Rak811()
    lora.sleep()
    if ctx.obj['VERBOSE']:
        click.echo('Sleeping')
    lora.close()


@cli.command(name='wake-up')
@click.pass_context
def wake_up(ctx):
    """Wake up."""
    lora = Rak811()
    lora.wake_up()
    if ctx.obj['VERBOSE']:
        click.echo('Alive!')
    lora.close()


@cli.command()
@click.argument(
    'reset_type',
    required=True,
    type=click.Choice(['module', 'lora'])
)
@click.pass_context
def reset(ctx, reset_type):
    """Reset Module or LoRaWan stack."""
    lora = Rak811()
    if reset_type == 'module':
        lora.reset(Reset.Module)
    else:
        lora.reset(Reset.LoRa)
    if ctx.obj['VERBOSE']:
        click.echo('{0} reset complete.'.format(
            'Module' if reset_type == 'module' else 'LoRa'))
    lora.close()


@cli.command()
@click.pass_context
def reload(ctx):
    """Set LoRaWan or LoRaP2P configurations to default."""
    lora = Rak811()
    lora.reload()
    if ctx.obj['VERBOSE']:
        click.echo('Configuration reloaded.')
    lora.close()


@cli.command()
@click.argument(
    'mode',
    required=False,
    type=click.Choice(['LoRaWan', 'LoRaP2P'], case_sensitive=False)
)
@click.pass_context
def mode(ctx, mode):
    """Get/Set mode to LoRaWan or LoRaP2P."""
    lora = Rak811()
    if mode is None:
        click.echo('LoRaWan' if lora.mode == Mode.LoRaWan else 'LoRaP2P')
    else:
        mode = mode.lower()
        if mode == 'lorawan':
            lora.mode = Mode.LoRaWan
        else:
            lora.mode = Mode.LoRaP2P
        if ctx.obj['VERBOSE']:
            click.echo('Mode set to {0}.'.format(
                'LoRaWan' if mode == 'lorawan' else 'LoRaP2P'))
    lora.close()


@cli.command()
@click.argument(
    'recv_ex',
    required=False,
    type=click.Choice(['enable', 'disable'])
)
@click.pass_context
def recv_ex(ctx, recv_ex):
    """RSSI & SNR report on receive."""
    lora = Rak811()
    if recv_ex is None:
        click.echo('Enabled' if lora.recv_ex == RecvEx.Enabled else 'Disabled')
    else:
        lora.recv_ex = (
            RecvEx.Enabled if recv_ex == 'enable' else RecvEx.Disabled
        )
        if ctx.obj['VERBOSE']:
            click.echo('RSSI & SNR report on receive {0}.'.format(
                'Enabled' if recv_ex == 'enable' else 'Disabled'))
    lora.close()


"""LoRaWan commands."""


@cli.command()
@click.argument(
    'band',
    required=False,
    type=click.Choice(
        ['EU868', 'US915', 'AU915', 'KR920', 'AS923', 'IN865'],
        case_sensitive=False
    )
)
@click.pass_context
def band(ctx, band):
    """Get/Set LoRaWan region."""
    lora = Rak811()
    if band is None:
        click.echo(lora.band)
    else:
        band = band.upper()
        lora.band = band
        if ctx.obj['VERBOSE']:
            click.echo('LoRaWan region set to {0}.'.format(band))
    lora.close()


@cli.command()
@click.argument(
    'key_values',
    metavar='KEY=VALUE...',
    required=True,
    type=KeyValueParamTypeLW(),
    nargs=-1
)
@click.pass_context
def set_config(ctx, key_values):
    """Set LoraWAN configuration.

    \b
    Arguments are specified as KEY=VALUE pairs, e.g.:
        set-config app_eui='APP_EUI' app_key='APP_KEY'
    """
    lora = Rak811()
    kv_args = dict(key_values)
    try:
        lora.set_config(**kv_args)
        if ctx.obj['VERBOSE']:
            click.echo('LoRaWan parameters set')
    except Rak811Error as e:
        print_exception(e)
    lora.close()


@cli.command()
@click.argument(
    'key',
    required=True,
    type=click.Choice(LW_CONFIG_KEYS)
)
@click.pass_context
def get_config(ctx, key):
    """Get LoraWan configuration."""
    lora = Rak811()
    try:
        click.echo(lora.get_config(key))
    except Rak811Error as e:
        print_exception(e)
    lora.close()


@cli.command()
@click.pass_context
def join_otaa(ctx):
    """Join the configured network in OTAA mode."""
    lora = Rak811()
    try:
        lora.join_otaa()
        if ctx.obj['VERBOSE']:
            click.echo('Joined in OTAA mode')
    except Rak811Error as e:
        print_exception(e)
    lora.close()


@cli.command()
@click.pass_context
def join_abp(ctx):
    """Join the configured network in ABP mode."""
    lora = Rak811()
    try:
        lora.join_abp()
        if ctx.obj['VERBOSE']:
            click.echo('Joined in ABP mode')
    except Rak811Error as e:
        print_exception(e)
    lora.close()


@cli.command()
@click.pass_context
def signal(ctx):
    """Get (RSSI,SNR) from latest received packet."""
    lora = Rak811()
    (rssi, snr) = lora.signal
    if ctx.obj['VERBOSE']:
        click.echo('RSSI: {0} - SNR: {1}'.format(rssi, snr))
    else:
        click.echo('{} {}'.format(rssi, snr))
    lora.close()


@cli.command()
@click.argument(
    'dr',
    required=False,
    type=click.INT
)
@click.pass_context
def dr(ctx, dr):
    """Get/Set next send data rate."""
    lora = Rak811()
    if dr is None:
        click.echo(lora.dr)
    else:
        try:
            lora.dr = dr
            if ctx.obj['VERBOSE']:
                click.echo('Data rate set to {0}.'.format(dr))
        except Rak811Error as e:
            print_exception(e)
    lora.close()


@cli.command()
@click.pass_context
def link_cnt(ctx):
    """Get up & downlink counters."""
    lora = Rak811()
    (uplink, downlink) = lora.link_cnt
    if ctx.obj['VERBOSE']:
        click.echo('Uplink: {0} - Downlink: {1}'.format(uplink, downlink))
    else:
        click.echo('{} {}'.format(uplink, downlink))
    lora.close()


@cli.command()
@click.pass_context
def abp_info(ctx):
    """Get ABP info.

    When using OTAA, returns the necessary info to re-join in ABP mode. The
    following tuple is returned: (NetworkID, DevAddr, Nwkskey, Appskey)
    """
    lora = Rak811()
    (nwk_id, dev_addr, nwks_key, apps_key) = lora.abp_info
    if ctx.obj['VERBOSE']:
        click.echo('NwkId: {}'.format(nwk_id))
        click.echo('DevAddr: {}'.format(dev_addr))
        click.echo('Nwkskey: {}'.format(nwks_key))
        click.echo('Appskey: {}'.format(apps_key))
    else:
        click.echo('{} {} {} {}'.format(nwk_id, dev_addr, nwks_key, apps_key))
    lora.close()


@cli.command()
@click.option(
    '-p', '--port',
    default=1,
    type=click.IntRange(1, 223),
    help='port number to use (1-223)'
)
@click.option(
    '--confirm',
    is_flag=True,
    help='regular or confirmed send'
)
@click.option(
    '--binary',
    is_flag=True,
    help='Data is binary (hex encoded)'
)
@click.argument(
    'data',
    required=True
)
@click.option(
    '--json',
    is_flag=True,
    help='Output downlink in JSON format'
)
@click.pass_context
def send(ctx, port, confirm, binary, data, json):
    """Send LoRaWan message and check for downlink."""
    if binary:
        try:
            data = bytes.fromhex(data)
        except ValueError:
            click.echo('Invalid binary data')
            return
    lora = Rak811()
    try:
        lora.send(data, confirm=confirm, port=port)
    except Rak811Error as e:
        print_exception(e)
        lora.close()
        return

    if ctx.obj['VERBOSE']:
        click.echo('Message sent.')
    if lora.nb_downlinks:
        downlink = lora.get_downlink()
        downlink['data'] = downlink['data'].hex()
        if json:
            click.echo(dumps(downlink, indent=4))
        elif ctx.obj['VERBOSE']:
            click.echo('Downlink received:')
            click.echo('Port: {}'.format(downlink['port']))
            if downlink['rssi']:
                click.echo('RSSI: {}'.format(downlink['rssi']))
                click.echo('SNR: {}'.format(downlink['snr']))
            click.echo('Data: {}'.format(downlink['data']))
        else:
            click.echo(downlink['data'])
    elif ctx.obj['VERBOSE']:
        click.echo('No downlink available.')
    lora.close()


@cli.command()
@click.argument(
    'key_values',
    metavar='KEY=VALUE...',
    required=False,
    type=KeyValueParamTypeP2P(),
    nargs=-1
)
@click.pass_context
def rf_config(ctx, key_values):
    """Get/Set LoraP2P configuration.

    \b
    Without argument, returns:
        frequency, sf, bw, cr, prlen, pwr

    \b
    Otherwise set rf_config, Arguments are specified as KEY=VALUE pairs:
        freq: frequency in MHz (860.000-929.900)
        sf: strength factor (6-12)
        bw: bandwidth (0:125KHz, 1:250KHz, 2:500KHz)
        cr: coding rate (1:4/5, 2:4/6, 3:4/7, 4:4/8)
        prlen: preamble length default (8-65535)
        pwr: Tx power (5-20)
    E.g.: rf-config freq=860.100 sf=7 pwr=16

    """
    lora = Rak811()
    config = dict(key_values)
    if config == {}:
        # No parameters: returns rc_config
        config = lora.rf_config
        if ctx.obj['VERBOSE']:
            click.echo('Frequency: {}'.format(config['freq']))
            click.echo('SF: {}'.format(config['sf']))
            click.echo('BW: {}'.format(config['bw']))
            click.echo('CR: {}'.format(config['cr']))
            click.echo('PrLen: {}'.format(config['prlen']))
            click.echo('Power: {}'.format(config['pwr']))
        else:
            click.echo('{} {} {} {} {} {}'.format(
                config['freq'], config['sf'], config['bw'], config['cr'],
                config['prlen'], config['pwr']
            ))
    else:
        # At least a parameter, set rc_config
        lora.rf_config = config
        if ctx.obj['VERBOSE']:
            click.echo('rf_config set: ' + ', '.join('{}={}'.format(k, v) for
                                                     k, v in config.items()))

    lora.close()


@cli.command()
@click.option(
    '--cnt',
    default=1,
    type=click.IntRange(1, 65535),
    help='tx counts (1-65535)'
)
@click.option(
    '--interval',
    default=60,
    type=click.IntRange(1, 3600),
    help=' tx interval (1-3600)'
)
@click.option(
    '--binary',
    is_flag=True,
    help='Data is binary (hex encoded)'
)
@click.argument(
    'data',
    required=True
)
@click.pass_context
def txc(ctx, cnt, interval, binary, data):
    """Send LoRaP2P message."""
    if binary:
        try:
            data = bytes.fromhex(data)
        except ValueError:
            click.echo('Invalid binary data')
            return
    lora = Rak811()
    try:
        lora.txc(data, cnt=cnt, interval=interval)
    except Rak811Error as e:
        print_exception(e)
        lora.close()
        return

    if ctx.obj['VERBOSE']:
        click.echo('Message sent.')
    lora.close()


@cli.command()
@click.pass_context
def rxc(ctx):
    """Set module in LoraP2P receive mode."""
    lora = Rak811()
    lora.rxc()
    if ctx.obj['VERBOSE']:
        click.echo('Module set in receive mode.')
    lora.close()


@cli.command()
@click.pass_context
def tx_stop(ctx):
    """Stop LoraP2P TX."""
    lora = Rak811()
    lora.tx_stop()
    if ctx.obj['VERBOSE']:
        click.echo('LoraP2P TX stopped.')
    lora.close()


@cli.command()
@click.pass_context
def rx_stop(ctx):
    """Stop LoraP2P RX."""
    lora = Rak811()
    lora.rx_stop()
    if ctx.obj['VERBOSE']:
        click.echo('LoraP2P RX stopped.')
    lora.close()


@cli.command()
@click.argument(
    'timeout',
    required=False,
    default=60,
    type=click.INT
)
@click.option(
    '--json',
    is_flag=True,
    help='Output message in JSON format'
)
@click.pass_context
def rx_get(ctx, timeout, json):
    """Get LoraP2P message."""
    lora = Rak811()
    lora.rx_get(timeout)
    if lora.nb_downlinks:
        rx = lora.get_downlink()
        rx['data'] = rx['data'].hex()
        if json:
            click.echo(dumps(rx, indent=4))
        elif ctx.obj['VERBOSE']:
            click.echo('Message received:')
            if rx['rssi']:
                click.echo('RSSI: {}'.format(rx['rssi']))
                click.echo('SNR: {}'.format(rx['snr']))
            click.echo('Data: {}'.format(rx['data']))
        else:
            click.echo(rx['data'])
    elif ctx.obj['VERBOSE']:
        click.echo('No message available.')
    lora.close()


@cli.command()
@click.pass_context
def radio_status(ctx):
    """Get radio statistics.

    Returns: TxSuccessCnt, TxErrCnt, RxSuccessCnt, RxTimeOutCnt, RxErrCnt,
    Rssi, Snr.
    """
    lora = Rak811()
    (
        tx_success_cnt, tx_err_cnt,
        rx_success_cnt, rx_timeout_cnt, rx_err_cnt,
        rssi, snr
    ) = lora.radio_status
    if ctx.obj['VERBOSE']:
        click.echo('TxSuccessCnt: {}'.format(tx_success_cnt))
        click.echo('TxErrCnt: {}'.format(tx_err_cnt))
        click.echo('RxSuccessCnt: {}'.format(rx_success_cnt))
        click.echo('RxTimeOutCnt: {}'.format(rx_timeout_cnt))
        click.echo('RxErrCnt: {}'.format(rx_err_cnt))
        click.echo('RSSI: {}'.format(rssi))
        click.echo('SNR: {}'.format(snr))
    else:
        click.echo('{} {} {} {} {} {} {}'.format(
            tx_success_cnt, tx_err_cnt,
            rx_success_cnt, rx_timeout_cnt, rx_err_cnt,
            rssi, snr
        ))
    lora.close()


@cli.command()
@click.pass_context
def clear_radio_status(ctx):
    """Clear radio statistics."""
    lora = Rak811()
    lora.clear_radio_status()
    if ctx.obj['VERBOSE']:
        click.echo('Radio statistics cleared.')
    lora.close()


if __name__ == '__main__':
    cli()



"""Units tests for the Rak811 class.

Rak811Serial is tested separately and therefore mocked in this suite.

RPi.GPIO is completely ignored as not available on all platforms, and its use
by the Rak811 class very limited.
"""
from mock import call, Mock, patch
from pytest import fixture, raises
# Ignore RPi.GPIO
p = patch.dict('sys.modules', {'RPi': Mock()})
p.start()
from rak811.rak811 import Mode, RecvEx, Reset  # noqa: E402
from rak811.rak811 import Rak811, Rak811EventError, \
    Rak811ResponseError  # noqa: E402
from rak811.rak811 import Rak811Serial  # noqa: E402
from rak811.rak811 import RESPONSE_TIMEOUT  # noqa: E402
from rak811.serial import Rak811TimeoutError  # noqa: E402


@fixture
def lora():
    """Instantiate Rak811 class though fixture."""
    with patch('rak811.rak811.Rak811Serial', autospec=True):
        rak811 = Rak811()
        yield rak811
        rak811.close()


@patch('rak811.rak811.Rak811Serial', autospec=True)
def test_instantiate_default(mock_serial):
    """Test that Rak811 can be instantiated.

    Check for basic initialisation and teardown of the RackSerial.
    """
    lora = Rak811()

    assert isinstance(lora._serial, Rak811Serial)
    mock_serial.assert_called_once_with(read_buffer_timeout=RESPONSE_TIMEOUT)
    lora.close()
    mock_serial.return_value.close.assert_called_once()


@patch('rak811.rak811.Rak811Serial', autospec=True)
def test_instantiate_params(mock_serial):
    """Test that Rak811 passes parameters passed to RackSerial."""
    port = '/dev/ttyAMA0'
    timeout = 5
    lora = Rak811(port=port, timeout=timeout)
    mock_serial.assert_called_once_with(
        read_buffer_timeout=RESPONSE_TIMEOUT,
        port=port,
        timeout=timeout
    )
    lora.close()


@patch('rak811.rak811.GPIO')
@patch('rak811.rak811.sleep')
def test_hard_reset(mock_sleep, mock_gpio, lora):
    """Test hard_reset().

    Test is a bit pointless, at least ensures it runs...
    """
    lora.hard_reset()
    mock_gpio.setup.assert_called_once()
    mock_gpio.cleanup.assert_not_called()


def test_int(lora):
    """Test _int function."""
    assert lora._int(1) == 1
    assert lora._int('1') == 1
    assert lora._int('Hello') == 'Hello'


def test_send_string(lora):
    """Test _send_string (passthrough to RackSerial)."""
    lora._send_string('Hello')
    lora._serial.send_string.assert_called_with('Hello')


def test_send_command(lora):
    """Test _send_command."""
    # Successful command
    lora._serial.receive.return_value = 'OK0'
    assert lora._send_command('dr') == '0'
    lora._serial.send_command.assert_called_with('dr')

    # Error
    lora._serial.receive.return_value = 'ERROR-1'
    with raises(Rak811ResponseError,
                match='-1'):
        lora._send_command('mode=2')

    # Unknown error
    lora._serial.receive.return_value = 'Unexpected'
    with raises(Rak811ResponseError,
                match='Unexpected'):
        lora._send_command('mode=2')

    # Events in response queue
    lora._serial.receive.side_effect = [
        'at+recv=2,0,0',
        'OK0'
    ]
    assert lora._send_command('dr') == '0'


def test_get_events(lora):
    """Test _get_events."""
    # Successful command
    lora._serial.receive.return_value = [
        'at+recv=2,0,0',
        'at+recv=0,1,0,0,1,55',
    ]
    events = lora._get_events()
    assert events.pop() == '0,1,0,0,1,55'
    assert events.pop() == '2,0,0'


"""AT command API.

For the sake of simplicity we mock Rak811.__send_command and
Rack811._get_events, as these have already been tested.
"""


@patch.object(Rak811, '_send_command', return_value='2.0.3.0')
def test_version(mock_send, lora):
    """Test version command."""
    assert lora.version == '2.0.3.0'
    mock_send.assert_called_once_with('version')


@patch.object(Rak811, '_send_command')
def test_sleep(mock_send, lora):
    """Test sleep command."""
    lora.sleep()
    mock_send.assert_called_once_with('sleep')


@patch.object(Rak811, '_get_events', return_value=['8,0,0'])
@patch.object(Rak811, '_send_string')
def test_wake_up(mock_send, mock_events, lora):
    """Test wake_up command."""
    lora.wake_up()
    mock_send.assert_called_once()
    mock_events.assert_called_once()


@patch.object(Rak811, 'hard_reset')
@patch.object(Rak811, '_send_command')
def test_reset_module(mock_send, mock_hard_reset, lora):
    """Test reset module command."""
    lora.reset(Reset.Module)
    mock_send.assert_called_once_with('reset=0')
    mock_hard_reset.assert_called_once()


@patch.object(Rak811, 'hard_reset')
@patch.object(Rak811, '_send_command')
def test_reset_lora(mock_send, mock_hard_reset, lora):
    """Test reset lora command."""
    lora.reset(Reset.LoRa)
    mock_send.assert_called_once_with('reset=1')
    mock_hard_reset.assert_not_called()


@patch.object(Rak811, '_send_command')
def test_reload(mock_send, lora):
    """Test reload command."""
    lora.reload()
    mock_send.assert_called_once_with('reload')


@patch.object(Rak811, '_send_command')
def test_set_mode(mock_send, lora):
    """Test mode setter."""
    lora.mode = Mode.LoRaWan
    mock_send.assert_called_once_with('mode=0')


@patch.object(Rak811, '_send_command', return_value='0')
def test_get_mode(mock_send, lora):
    """Test mode getter."""
    assert lora.mode == 0
    mock_send.assert_called_once_with('mode')


@patch.object(Rak811, '_send_command')
def test_set_recv_ex(mock_send, lora):
    """Test recv_ex setter."""
    lora.recv_ex = RecvEx.Disabled
    mock_send.assert_called_once_with('recv_ex=1')


@patch.object(Rak811, '_send_command', return_value='1')
def test_get_recv_ex(mock_send, lora):
    """Test recv_ex getter."""
    assert lora.recv_ex == 1
    mock_send.assert_called_once_with('recv_ex')


@patch.object(Rak811, '_send_command')
def test_set_config(mock_send, lora):
    """Test config setter."""
    lora.set_config(adr='on', dr=5)
    assert len(mock_send.mock_calls) == 1
    assert mock_send.mock_calls[0] in (
        call('set_config=adr:on&dr:5'),
        call('set_config=dr:5&adr:on')
    )


@patch.object(Rak811, '_send_command', return_value='1')
def test_get_config(mock_send, lora):
    """Test config getter."""
    assert lora.get_config('dr') == '1'
    mock_send.assert_called_once_with('get_config=dr')


@patch.object(Rak811, '_send_command')
def test_set_band(mock_send, lora):
    """Test band setter."""
    lora.band = 'EU868'
    mock_send.assert_called_once_with('band=EU868')


@patch.object(Rak811, '_send_command', return_value='EU868')
def test_get_band(mock_send, lora):
    """Test band getter."""
    assert lora.band == 'EU868'
    mock_send.assert_called_once_with('band')


@patch.object(Rak811, '_send_command')
def test_join_abp(mock_send, lora):
    """Test join_abp command."""
    lora.join_abp()
    mock_send.assert_called_once_with('join=abp')


@patch.object(Rak811, '_get_events', return_value=['3,0,0'])
@patch.object(Rak811, '_send_command')
def test_join_otaa_success(mock_send, mock_events, lora):
    """Test join_abp command, successful."""
    lora.join_otaa()
    mock_send.assert_called_once_with('join=otaa')
    mock_events.assert_called_once()


@patch.object(Rak811, '_get_events', return_value=['4,0,0'])
@patch.object(Rak811, '_send_command')
def test_join_otaa_failure(mock_send, mock_events, lora):
    """Test join_abp command, failure."""
    with raises(Rak811EventError,
                match='4'):
        lora.join_otaa()
    mock_send.assert_called_once_with('join=otaa')
    mock_events.assert_called_once()


@patch.object(Rak811, '_send_command', return_value='-30,26')
def test_signal(mock_send, lora):
    """Test signal command."""
    assert lora.signal == (-30, 26)
    mock_send.assert_called_once_with('signal')


@patch.object(Rak811, '_send_command')
def test_set_dr(mock_send, lora):
    """Test dr setter."""
    lora.dr = 5
    mock_send.assert_called_once_with('dr=5')


@patch.object(Rak811, '_send_command', return_value='5')
def test_get_dr(mock_send, lora):
    """Test dr getter."""
    assert lora.dr == 5
    mock_send.assert_called_once_with('dr')


@patch.object(Rak811, '_send_command')
def test_set_link_cnt(mock_send, lora):
    """Test link_cnt setter."""
    lora.link_cnt = (15, 2)
    mock_send.assert_called_once_with('link_cnt=15,2')


@patch.object(Rak811, '_send_command', return_value='15,2')
def test_get_link_cnt(mock_send, lora):
    """Test link_cnt getter."""
    assert lora.link_cnt == (15, 2)
    mock_send.assert_called_once_with('link_cnt')


@patch.object(Rak811, '_send_command', return_value=(
    # cSpell:disable
    '13,'
    '26dddddd,'
    '9annnnnnnnnnnnnnnnnnnnnnnnnnnnnn,'
    '0baaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
    # cSpell:enable
))
def test_abp_info(mock_send, lora):
    """Test abp_info command."""
    assert lora.abp_info == (
        # cSpell:disable
        '13',
        '26dddddd',
        '9annnnnnnnnnnnnnnnnnnnnnnnnnnnnn',
        '0baaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
        # cSpell:enable
    )
    mock_send.assert_called_once_with('abp_info')


@patch.object(Rak811, '_get_events', return_value=['2,0,0'])
@patch.object(Rak811, '_send_command')
def test_get_send_unconfirmed(mock_send, mock_events, lora):
    """Test send, unconfirmed."""
    lora.send('Hello')
    mock_send.assert_called_once_with('send=0,1,48656c6c6f')
    mock_events.assert_called_once()
    assert lora.nb_downlinks == 0


@patch.object(Rak811, '_get_events', return_value=['1,0,0'])
@patch.object(Rak811, '_send_command')
def test_get_send_confirmed(mock_send, mock_events, lora):
    """Test send, confirmed."""
    lora.send('Hello', port=2, confirm=True)
    mock_send.assert_called_once_with('send=1,2,48656c6c6f')
    mock_events.assert_called_once()
    assert lora.nb_downlinks == 0


@patch.object(Rak811, '_get_events', return_value=['2,0,0'])
@patch.object(Rak811, '_send_command')
def test_get_send_binary(mock_send, mock_events, lora):
    """Test send, binary."""
    lora.send(bytes.fromhex('01020211'))
    mock_send.assert_called_once_with('send=0,1,01020211')
    mock_events.assert_called_once()


@patch.object(Rak811, '_get_events', return_value=['5,0,0'])
@patch.object(Rak811, '_send_command')
def test_get_send_fail(mock_send, mock_events, lora):
    """Test send, fail."""
    with raises(Rak811EventError,
                match='5'):
        lora.send(bytes.fromhex('01020211'))
    mock_send.assert_called_once_with('send=0,1,01020211')
    mock_events.assert_called_once()


@patch.object(Rak811, '_get_events', return_value=[
    '2,0,0',
    '0,11,-34,27,4,65666768',
])
@patch.object(Rak811, '_send_command')
def test_get_send_receive(mock_send, mock_events, lora):
    """Test send and receive."""
    lora.send(bytes.fromhex('01020211'))
    mock_send.assert_called_once_with('send=0,1,01020211')
    mock_events.assert_called_once()
    assert lora.nb_downlinks == 1
    assert lora.get_downlink() == {
        'port': 11,
        'rssi': -34,
        'snr': 27,
        'len': 4,
        'data': bytes.fromhex('65666768'),
    }


@patch.object(Rak811, '_get_events', return_value=[
    '2,0,0',
    '0,11,4,65666768',
])
@patch.object(Rak811, '_send_command')
def test_get_send_receive_no_recv_ex(mock_send, mock_events, lora):
    """Test send and receive, recv_ex disabled."""
    lora.send(bytes.fromhex('01020211'))
    mock_send.assert_called_once_with('send=0,1,01020211')
    mock_events.assert_called_once()
    assert lora.nb_downlinks == 1
    assert lora.get_downlink() == {
        'port': 11,
        'rssi': 0,
        'snr': 0,
        'len': 4,
        'data': bytes.fromhex('65666768'),
    }


@patch.object(Rak811, '_send_command', return_value='868700000,7,1,3,4,5')
def test_get_rf_config(mock_send, lora):
    """Test config getter."""
    assert lora.rf_config == {
        'freq': 868.700,
        'sf': 7,
        'bw': 1,
        'cr': 3,
        'prlen': 4,
        'pwr': 5
    }
    mock_send.assert_called_once_with('rf_config')


@patch.object(Rak811, '_get_rf_config', return_value={
    'freq': 868.100,
    'sf': 12,
    'bw': 0,
    'cr': 1,
    'prlen': 8,
    'pwr': 20
})
@patch.object(Rak811, '_send_command')
def test_set_rf_config_default(mock_send, mock_rf_config, lora):
    """Test RF config setter (No parameters)."""
    lora.rf_config = {}
    mock_rf_config.assert_called_once()
    mock_send.assert_called_once_with('rf_config=868100000,12,0,1,8,20')


@patch.object(Rak811, '_get_rf_config', return_value={
    'freq': 868.100,
    'sf': 12,
    'bw': 1,
    'cr': 1,
    'prlen': 8,
    'pwr': 20
})
@patch.object(Rak811, '_send_command')
def test_set_rf_config_partial(mock_send, mock_rf_config, lora):
    """Test RF config setter (Partial input)."""
    lora.rf_config = {
        'freq': 868.700,
        'sf': 7,
        'bw': 0
    }
    mock_rf_config.assert_called_once()
    mock_send.assert_called_once_with('rf_config=868700000,7,0,1,8,20')


@patch.object(Rak811, '_get_rf_config', return_value={
    'freq': 868.100,
    'sf': 12,
    'bw': 0,
    'cr': 1,
    'prlen': 8,
    'pwr': 20
})
@patch.object(Rak811, '_send_command')
def test_set_rf_config_complete(mock_send, mock_rf_config, lora):
    """Test RF config setter."""
    lora.rf_config = {
        'freq': 868.700,
        'sf': 7,
        'bw': 1,
        'cr': 3,
        'prlen': 4,
        'pwr': 5
    }
    mock_rf_config.assert_called_once()
    mock_send.assert_called_once_with('rf_config=868700000,7,1,3,4,5')


@patch.object(Rak811, '_get_events', return_value=['9,0,0'])
@patch.object(Rak811, '_send_command')
def test_txc(mock_send, mock_events, lora):
    """Test LoraP2P send."""
    lora.txc('Hello')
    mock_send.assert_called_once_with('txc=1,60000,48656c6c6f')
    mock_events.assert_called_once()


@patch.object(Rak811, '_get_events', return_value=['5,0,0'])
@patch.object(Rak811, '_send_command')
def test_txc_error(mock_send, mock_events, lora):
    """Test LoraP2P send with error."""
    with raises(Rak811EventError,
                match='5'):
        lora.txc('Hello')
    mock_send.assert_called_once_with('txc=1,60000,48656c6c6f')
    mock_events.assert_called_once()


@patch.object(Rak811, '_send_command')
def test_rxc(mock_send, lora):
    """Test LoraP2P RXC."""
    lora.rxc()
    mock_send.assert_called_once_with('rxc=1')


@patch.object(Rak811, '_send_command')
def test_tx_stop(mock_send, lora):
    """Test LoraP2P tx stop."""
    lora.tx_stop()
    mock_send.assert_called_once_with('tx_stop')


@patch.object(Rak811, '_send_command')
def test_rx_stop(mock_send, lora):
    """Test LoraP2P rx stop."""
    lora.rx_stop()
    mock_send.assert_called_once_with('rx_stop')


@patch.object(Rak811, '_get_events', return_value=[
    '0,0,4,65666768',
])
def test_rx_get(mock_events, lora):
    """Test rx_get."""
    lora.rx_get(10)
    mock_events.assert_called_once()
    assert lora.nb_downlinks == 1
    assert lora.get_downlink() == {
        'port': 0,
        'rssi': 0,
        'snr': 0,
        'len': 4,
        'data': bytes.fromhex('65666768'),
    }


@patch.object(Rak811, '_get_events')
def test_rx_get_no_data(mock_events, lora):
    """Test rx_get with no data."""
    mock_events.side_effect = Rak811TimeoutError()
    lora.rx_get(10)
    mock_events.assert_called_once()
    assert lora.nb_downlinks == 0
    assert lora.get_downlink() is None


@patch.object(Rak811, '_send_command', return_value=('8,0,1,0,0,-48,28'))
def test_radio_status(mock_send, lora):
    """Test radio_status command."""
    assert lora.radio_status == (8, 0, 1, 0, 0, -48, 28)
    mock_send.assert_called_once_with('status')


@patch.object(Rak811, '_send_command')
def test_clear_radio_status(mock_send, lora):
    """Test clear_radio_status command."""
    lora.clear_radio_status()
    mock_send.assert_called_once_with('status=0')



#!/usr/bin/env python3
"""RAK811 API demo.

Simple demo to illustrate API usage

Copyright 2019, 2021 Philippe Vanhaesendonck

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

SPDX-License-Identifier: Apache-2.0
"""
from sys import exit
from time import sleep

from rak811.rak811 import Mode, RecvEx, Reset
from rak811.rak811 import Rak811, Rak811ResponseError
from ttn_secrets import APP_EUI, APP_KEY

config_keys = ('dev_addr', 'dev_eui', 'app_eui', 'app_key', 'nwks_key',
               'apps_key', 'tx_power', 'pwr_level', 'adr', 'dr', 'public_net',
               'rx_delay1', 'ch_list', 'ch_mask', 'max_chs', 'rx2',
               'join_cnt', 'nbtrans', 'retrans', 'class', 'duty')

print('Instanciate class')
lora = Rak811()

print('Hard reset board')
lora.hard_reset()

# System commands
print('Version', lora.version)

print('Sleeping')
lora.sleep()
sleep(5)
lora.wake_up()
print('Awake')

print('Reset Module')
lora.reset(Reset.Module)

print('Reset LoRa')
lora.reset(Reset.LoRa)

print('Reload')
lora.reload()

for mode in (Mode.LoRaP2P, Mode.LoRaWan):
    lora.mode = mode
    print('Mode', lora.mode)

for band in ('US915', 'EU868'):
    lora.band = band
    print('Band', lora.band)

print("Configure module")
lora.set_config(app_eui=APP_EUI,
                app_key=APP_KEY)
print("Module configuration:")
for config_key in config_keys:
    try:
        print('>>', config_key, lora.get_config(config_key))
    except Rak811ResponseError as e:
        print('>>', config_key, e.errno, e.strerror)

for recv_ex in (RecvEx.Disabled, RecvEx.Enabled):
    lora.recv_ex = recv_ex
    print('Recv ex', lora.recv_ex)

print('Join')
lora.join_otaa()

print('DR', lora.dr)
lora.dr = 5
print('DR', lora.dr)

print('Signal', lora.signal)

print('Send string')
lora.send('Hello')

print('Signal', lora.signal)

print('Link counter', lora.link_cnt)

print('ABP Info', lora.abp_info)

print('Send Confirmed (Cayenne LPP format)')
lora.send(bytes.fromhex('016700F0'), port=11, confirm=True)

print('Check for downlink messages')
while lora.nb_downlinks:
    print('>>', lora.get_downlink())

lora.close()
exit(0)



"""Interface with the RAK811 module (Firmware V2.0).

Copyright 2019, 2021 Philippe Vanhaesendonck

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

SPDX-License-Identifier: Apache-2.0
"""
from binascii import hexlify
from enum import IntEnum
from time import sleep

from RPi import GPIO

from .exception import Rak811Error
from .serial import Rak811Serial, Rak811TimeoutError

RESET_BCM_PORT = 17
RESET_DELAY = 0.01
RESET_POST = 2
RESPONSE_OK = 'OK'
RESPONSE_ERROR = 'ERROR'
RESPONSE_EVENT = 'at+recv='

# Timeout for response and events
# The RAK811 typically respond in less than 1.5 seconds
RESPONSE_TIMEOUT = 5
# Event wait time strongly depends on duty cycle, when sending often at high SF
# the module will wait to respect the duty cycle.
# In normal operation, 5 minutes should be more than enough.
EVENT_TIMEOUT = 5 * 60


# RAK811 error codes and associated messages
class ErrorCode(IntEnum):
    """AT commands error codes."""

    ARG_ERR = -1
    ARG_NOT_FIND = -2
    JOIN_ABP_ERR = -3
    JOIN_OTAA_ERR = -4
    NOT_JOIN = -5
    MAC_BUSY_ERR = -6
    TX_ERR = -7
    INTER_ERR = -8
    WR_CFG_ERR = -11
    RD_CFG_ERR = -12
    TX_LEN_LIMIT_ERR = -13
    UNKNOWN_ERR = -20


ERROR_MESSAGE = {
    ErrorCode.ARG_ERR: 'Invalid argument',
    ErrorCode.ARG_NOT_FIND: 'Argument not found',
    ErrorCode.JOIN_ABP_ERR: 'ABP join error',
    ErrorCode.JOIN_OTAA_ERR: 'OTAA join error',
    ErrorCode.NOT_JOIN: 'Not joined',
    ErrorCode.MAC_BUSY_ERR: 'MAC busy',
    ErrorCode.TX_ERR: 'Transmit error',
    ErrorCode.INTER_ERR: 'Inter error',
    ErrorCode.WR_CFG_ERR: 'Write configuration error',
    ErrorCode.RD_CFG_ERR: 'Read configuration Error',
    ErrorCode.TX_LEN_LIMIT_ERR: 'Transmit len limit error',
    ErrorCode.UNKNOWN_ERR: 'Unknown error',
}


class EventCode(IntEnum):
    """AT commands event codes."""

    RECV_DATA = 0
    TX_COMFIRMED = 1
    TX_UNCONFIRMED = 2
    JOINED_SUCCESS = 3
    JOINED_FAILED = 4
    TX_TIMEOUT = 5
    RX2_TIMEOUT = 6
    DOWNLINK_REPEATED = 7
    WAKE_UP = 8
    P2PTX_COMPLETE = 9
    UNKNOWN = 100


EVENT_MESSAGE = {
    EventCode.RECV_DATA: 'Received data',
    EventCode.TX_COMFIRMED: 'Tx confirmed',
    EventCode.TX_UNCONFIRMED: 'Tx unconfirmed',
    EventCode.JOINED_SUCCESS: 'Join succeeded',
    EventCode.JOINED_FAILED: 'Join failed',
    EventCode.TX_TIMEOUT: 'Tx timeout',
    EventCode.RX2_TIMEOUT: 'Rx2 timeout',
    EventCode.DOWNLINK_REPEATED: 'Downlink repeated',
    EventCode.WAKE_UP: 'Wake up',
    EventCode.P2PTX_COMPLETE: 'P2P tx complete',
    EventCode.UNKNOWN: 'Unknown',
}


class Mode(IntEnum):
    """Module operation mode (LoRaWan or Peer to Peer)."""

    LoRaWan = 0
    LoRaP2P = 1


class Reset(IntEnum):
    """Module reset type."""

    Module = 0
    LoRa = 1


class RecvEx(IntEnum):
    """Receive RSSI/SNR data with downlink packets."""

    Enabled = 0
    Disabled = 1


class Rak811ResponseError(Rak811Error):
    """Exception raised by response from the module.

    Attributes:
        errno -- as returned by the module
        strerror -- textual representation

    """

    def __init__(self, code):
        """Just assign return codes."""
        try:
            self.errno = int(code)
        except ValueError:
            self.errno = code

        if self.errno in ERROR_MESSAGE:
            self.strerror = ERROR_MESSAGE[self.errno]
        else:
            self.strerror = ERROR_MESSAGE[ErrorCode.UNKNOWN_ERR]
        super().__init__(('[Errno {}] {}').format(self.errno, self.strerror))


class Rak811EventError(Rak811Error):
    """Exception raised by module events.

    Attributes:
        errno  -- as returned by the module
        strerror -- textual representation

    """

    def __init__(self, status):
        """Just assign return status."""
        try:
            self.errno = int(status)
        except ValueError:
            self.errno = status

        if self.errno in EVENT_MESSAGE:
            self.strerror = EVENT_MESSAGE[self.errno]
        else:
            self.strerror = EVENT_MESSAGE[EventCode.UNKNOWN]
        super().__init__(('[Errno {}] {}').format(self.errno, self.strerror))


class Rak811(object):
    """Main class."""

    def __init__(self, **kwargs):
        """Initialise class.

        The serial port is immediately opened and flushed.
        All parameters are optional and passed to RackSerial.
        """
        read_buffer_timeout = kwargs.pop('response_timeout', RESPONSE_TIMEOUT)
        self._event_timeout = kwargs.pop('event_timeout', EVENT_TIMEOUT)
        self._serial = Rak811Serial(
            read_buffer_timeout=read_buffer_timeout,
            **kwargs
        )
        self._downlink = []

    def close(self):
        """Terminates session.

        Terminates read thread and close serial port.
        """
        self._serial.close()

    def hard_reset(self):
        """Hard reset of the RAK811 module.

        Hard reset should not be required in normal operation. It needs to be
        issued once after host boot, or module restart.
        Note that we do not cleanup() as the reset port should stay high (it is
        configured that way at boot time).
        """
        GPIO.setwarnings(False)
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(RESET_BCM_PORT, GPIO.OUT)
        GPIO.output(RESET_BCM_PORT, GPIO.LOW)
        sleep(RESET_DELAY)
        GPIO.output(RESET_BCM_PORT, GPIO.HIGH)
        sleep(RESET_POST)

    def _int(self, i):
        """Attempt int conversion."""
        try:
            i = int(i)
        except ValueError:
            pass
        return i

    def _send_string(self, string):
        """Send string to the RAK811 module."""
        self._serial.send_string(string)

    def _send_command(self, command):
        """Send AT command to the RAK811 module and return the response.

        Rak811ResponseError exception is raised if the command returns an
        error.
        This is a "blocking" call: if the module does not respond
        Rack811TimeoutError will be raised.
        """
        self._serial.send_command(command)
        response = self._serial.receive()

        # Ignore events received while waiting on command feedback
        while response.startswith(RESPONSE_EVENT):
            response = self._serial.receive()

        if response.startswith(RESPONSE_OK):
            response = response[len(RESPONSE_OK):]
        elif response.startswith(RESPONSE_ERROR):
            raise Rak811ResponseError(response[len(RESPONSE_ERROR):])
        else:
            raise Rak811ResponseError(response)

        return response

    def _get_events(self, timeout=None):
        """Get events from the RAK811 module.

        This is a "blocking" call: it will either return a list of events or
        raise a Rack811TimeoutError.
        """
        if timeout is None:
            timeout = self._event_timeout

        return [i[len(RESPONSE_EVENT):] for i in
                self._serial.receive(single=False, timeout=timeout)]

    """System commands."""

    @property
    def version(self):
        """Get module version."""
        return(self._send_command('version'))

    def sleep(self):
        """Enter sleep mode."""
        self._send_command('sleep')

    def wake_up(self):
        """Wake up the RAK811 module.

        We just send a character and wait for an event as response
        """
        self._send_string('*')
        self._get_events()

    def reset(self, mode):
        """Reset Module or LoRaWan stack.

        Note that reset(Reset.Module) will restart the module which will wait
        for an hardware reset to start.
        """
        self._send_command('reset={0}'.format(mode))
        if mode == Reset.Module:
            self.hard_reset()

    def reload(self):
        """Set LoraWan or LoraP2P configurations to default."""
        self._send_command('reload')

    @property
    def mode(self):
        """Get module Mode (LoRaWan or LoRaP2P)."""
        return(self._int(self._send_command('mode')))

    @mode.setter
    def mode(self, value):
        """Set module in LoRaWan or LoRaP2P Mode."""
        self._send_command('mode={0}'.format(value))

    @property
    def recv_ex(self):
        """Get RSSI & SNR report on receive flag (Enabled/Disabled)."""
        return(self._int(self._send_command('recv_ex')))

    @recv_ex.setter
    def recv_ex(self, value):
        """Set RSSI & SNR report on receive flag (Enabled/Disabled)."""
        self._send_command('recv_ex={0}'.format(value))

    """ LoRaWan commands."""

    def set_config(self, **kwargs):
        """Set LoraWan configuration.

        Parameters are specified a key/value pairs, and are passed directly to
        the module.

        E.g.: set_config(app_eui='0000000000000000',
                         app_key='00000000000000000000000000000000',
                         adr='on')

        The module saves relevant data in EEPROM, it is not necessary to set
        values for each session.

        The following parameters are accepted by the module:
            dev_addr: device address (4 bytes hex number)
            dev_eui: device EUI (8 bytes hex number, default derived from the
                MCU UUID
            app_eui: app EUI (8 bytes hex number)
            app_key: app key (16 bytes hex number)
            apps_key: application session key (16 bytes hex number)
            nwks_key: network session key (16 bytes hex number)
            tx_power: transmit power (in dBm -- deprecated, use pwr_level)
            pwr_level: transmit power (0-7 for EU868, region specific)
            adr: adr flag (on/off)
            dr: data rate (0-5 for EU868, region specific)
            public_net: public_net flag (on/off)
            rx_delay1: rx1 delay (0-65535 milliseconds)
            ch_list: channel list, see RAK documentation
            ch_mask: channel mask, see RAK documentation
            max_chs: max channels used in the region (read-only)
            rx2: rx2 data rate and frequency
            join_cnt: join count for OTAA joins (number, region specific)
            nbtrans: number of transmissions for unconfirmed uplink message
                (1-15, default 1)
            retrans: number of retransmissions for confirmed uplink message
                (1-255, default 8)
            class: LoRa class (0: A, 2: C)
            duty: respect duty cycle flag (on/off)
        """
        self._send_command('set_config='
                           + '&'.join([':'.join(str(val) for val in kv)
                                       for kv in kwargs.items()]))

    def get_config(self, key):
        """Get LoraWan configuration from EEPROM.

        The parameter must be a key from the above list.

        Note: get_config returns always strings, no integer do avoid unwanted
        conversion for keys.
        """
        return self._send_command('get_config={0}'.format(key))

    @property
    def band(self):
        """Get LoRaWan region.

        Region is one of: EU868, US915, AU915, KR920, AS923, IN865.
        """
        return(self._send_command('band'))

    @band.setter
    def band(self, region):
        """Set LoRaWan region.

        Region must be one of: EU868, US915, AU915, KR920, AS923, IN865.
        """
        self._send_command('band={0}'.format(region))

    def join_abp(self):
        """Join the configured network in ABP mode.

        ABP requires the following parameters to be set prior join:
            - dev_addr
            - nwks_key
            - apps_key
        """
        self._send_command('join=abp')

    def join_otaa(self):
        """Join the configured network in OTAA mode.

        OTAA requires the following parameters to be set prior join:
            - dev_eui
            - app_eui
            - app_key

        This call is "blocking", it will return only after the join completes.
        The following exceptions can be raised:
            - Rak811TimeoutError: join didn't succeed in time
            - Rak811EventError: join failed
        """
        self._send_command('join=otaa')
        # Waiting join completion
        for event in self._get_events():
            status = event.split(',')[0]
            status = self._int(status)
            if status != EventCode.JOINED_SUCCESS:
                raise Rak811EventError(status)

    @property
    def signal(self):
        """Get (RSSI,SNR) from latest received packet."""
        return(tuple(self._int(i)
                     for i in self._send_command('signal').split(',')))

    @property
    def dr(self):
        """Get next send data rate."""
        return(self._int(self._send_command('dr')))

    @dr.setter
    def dr(self, value):
        """Set next send data rate."""
        self._send_command('dr={0}'.format(value))

    @property
    def link_cnt(self):
        """Get up & downlink counters.

        Counters are 32 bits integers in decimal format.
        """
        return(tuple(self._int(i)
               for i in self._send_command('link_cnt').split(',')))

    @link_cnt.setter
    def link_cnt(self, value):
        """Set up & downlink counters.

        Counters are 32 bits integers in decimal format.
        """
        self._send_command('link_cnt=' + ','.join(str(val) for val in value))

    @property
    def abp_info(self):
        """Get ABP info.

        When using OTAA, returns the necessary info to re-join in ABP mode. The
        following tuple is returned:
            (NetworkID, DevAddr, Nwkskey, Appskey)
        """
        return(tuple(self._send_command('abp_info').split(',')))

    def _add_downlink(self, event):
        """Add message to the downlink list.

        Event is a list: (<port>[,<rssi>][,<snr>],<len>[,<data>])
        """
        r_port = self._int(event.pop(0))
        if len(event) > 2:
            r_rssi = self._int(event.pop(0))
            r_snr = self._int(event.pop(0))
        else:
            r_rssi = 0
            r_snr = 0
        r_len = self._int(event.pop(0))
        if r_len > 0:
            try:
                r_data = bytes.fromhex(event[0])
            except ValueError:
                r_data = ''
        else:
            r_data = ''
        self._downlink.append(
            {
                'port': r_port,
                'rssi': r_rssi,
                'snr': r_snr,
                'len': r_len,
                'data': r_data,
            }
        )

    def _process_events(self, timeout=None):
        """Process module event queue.

        Process event queue looking for incoming (downlink) messages. Raise
        errors when unexpected events are encountered.

        Parameter:
            timeout: maximum time to wait for event

        """
        events = self._get_events(timeout)
        # Check for downlink
        for event in events:
            # Format: <status >,<port>[,<rssi>][,<snr>],<len>[,<data>]
            event_items = event.split(',')
            status = event_items.pop(0)
            status = self._int(status)
            if status == EventCode.RECV_DATA:
                self._add_downlink(event_items)
        # Check for errors
        for event in events:
            status = event.split(',')[0]
            status = self._int(status)
            if status not in (EventCode.RECV_DATA,
                              EventCode.TX_COMFIRMED,
                              EventCode.TX_UNCONFIRMED):
                raise Rak811EventError(status)

    def send(self, data, confirm=False, port=1):
        """Send LoRaWan message.

        Parameters:
            data: data to be sent. If the datatype is bytes it will be send
                as such. Strings will be converted to bytes.
            confirm: regular or confirmed send.
            port: port number to use (1-223)

        """
        if type(data) is not bytes:
            data = (bytes)(data, 'utf-8')
        data = hexlify(data).decode('ascii')

        self._send_command('send=' + ','.join((
            ('1' if confirm else '0'),
            str(port),
            data
        )))

        # Process events
        self._process_events()

    @property
    def nb_downlinks(self):
        """Get the number of downlink messages in the receive buffer."""
        return len(self._downlink)

    def get_downlink(self):
        """Get a downlink message from the receive buffer.

        Returns a dictionary with the following keys:
            port: port number
            rssi: RSSI (0 if recv_ex was disabled)
            snr: SNR (0 if recv_ex was disabled)
            len: data length
            data: data itself
        """
        if len(self._downlink) == 0:
            return None
        else:
            return self._downlink.pop(0)

    """ LoraP2P commands."""

    def _get_rf_config(self):
        """Get LoraP2P configuration.

        Return a dictionary.
        """
        config = tuple(self._int(i)
                       for i in self._send_command('rf_config').split(','))
        return {
            'freq': config[0] / 1000 / 1000,
            'sf': config[1],
            'bw': config[2],
            'cr': config[3],
            'prlen': config[4],
            'pwr': config[5]
        }

    @property
    def rf_config(self):
        """Get LoraP2P configuration.

        Return a dictionary.
        """
        return self._get_rf_config()

    @rf_config.setter
    def rf_config(self, config):
        """Set LoraWan P2P RF configuration parameters.

        Parameters are specified a key/value pairs, default are used for
        missing parameters.

        E.g.:
            rf_config = {
                'freq': 868.700,
                'sf': 7,
                'bw': 0
            }

        The module saves parameters to flash, it is not necessary to set
        values for each session.

        The following parameters can be set; only specified parameters are
        changed, others are kept to their previous value. Values between
        parentheses are RAK defaults.
            freq: frequency in Mhz, range 860.000-929.900 Mhz (868.100)
            sf: spread factor, range 6-12 (12)
            bw: band width, values 0:125KHz, 1:250KHz, 2:500KHz (0)
            cr: coding rate, values 1:4/5, 2:4/6, 3:4/7, 4:4/8 (1)
            prlen: preamble len, range 8-65536 (8)
            pwr: transmit power, range 5,20 (20)
        """
        base_config = self._get_rf_config()
        base_config.update(config)
        self._send_command(
            'rf_config={0},{1},{2},{3},{4},{5}'.format(
                int(base_config['freq'] * 1000 * 1000),
                base_config['sf'],
                base_config['bw'],
                base_config['cr'],
                base_config['prlen'],
                base_config['pwr']
            )
        )

    def txc(self, data, cnt=1, interval=60):
        """Send LoraP2P message.

        Send data using the pre-set RF parameters.
        For RF testing cnt can be specified to send data multiple time.
        The module will stop sending messages after cnt messages or when it
        receives a tx_stop command.
        The method returns after all messages have been sent.

        Parameters:
            data: data to be sent. If the datatype is bytes it will be send
                as such. Strings will be converted to bytes.
            cnt: send message cnt times
            interval: when sending multiple times, interval in seconds
            between each message.

        """
        if type(data) is not bytes:
            data = (bytes)(data, 'utf-8')
        data = hexlify(data).decode('ascii')

        self._send_command('txc=' + ','.join((
            str(cnt),
            str(interval * 1000),
            data
        )))

        # Process events
        events = self._get_events((cnt * (interval + 10)) - interval)
        # Check for errors
        for event in events:
            status = event.split(',')[0]
            status = self._int(status)
            if status != EventCode.P2PTX_COMPLETE:
                raise Rak811EventError(status)

    def rxc(self, report_en=1):
        """Set module in LoraP2P receive mode.

        Module is put in receive mode until an rx_stop command is issued.
        Method will return immediately after the command is acknowledged (it
        does not wait for data)

        Parameter:
            report_en: set to 1 by default. Can be set to 0 for RF testing
            (documentation is not clear about this)
        """
        self._send_command('rxc=' + str(report_en))

    def tx_stop(self):
        """Stop LoraP2P TX.

        Stop LoraP2P transmission; radio will switch to sleep mode.
        """
        self._send_command('tx_stop')

    def rx_stop(self):
        """Stop LoraP2P RX.

        Stop LoraP2P reception; radio will switch to sleep mode.
        """
        self._send_command('rx_stop')

    def rx_get(self, timeout):
        """Get LoraP2P message.

        This is a blocking call: wait until we receive a message or we reach a
        timeout.

        The downlink receive buffer is populated, actual data is retrieved with
        get_downlink().
        """
        try:
            self._process_events(timeout)
        except Rak811TimeoutError:
            pass

    """Radio commands."""

    @property
    def radio_status(self):
        """Get radio statistics.

        Return a tuple: (TxSuccessCnt, TxErrCnt, RxSuccessCnt, RxTimeOutCnt,
                         RxErrCnt, Rssi, Snr)
        """
        return(tuple(self._int(i)
               for i in self._send_command('status').split(',')))

    def clear_radio_status(self):
        """Clear radio statistics."""
        self._send_command('status=0')



#!/usr/bin/env python3
"""RAK811 OTAA demo.

Minimalistic OTAA demo

Copyright 2019, 2021 Philippe Vanhaesendonck

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

SPDX-License-Identifier: Apache-2.0
"""
from random import randint
from sys import exit
from time import sleep

from rak811.rak811 import Mode, Rak811
from ttn_secrets import APP_EUI, APP_KEY

lora = Rak811()

# Most of the setup should happen only once...
print('Setup')
lora.hard_reset()
lora.mode = Mode.LoRaWan
lora.band = 'EU868'
lora.set_config(app_eui=APP_EUI,
                app_key=APP_KEY)

print('Joining')
lora.join_otaa()
# Note that DR is different from SF and depends on the region
# See: https://docs.exploratory.engineering/lora/dr_sf/
# Set Data Rate to 5 which is SF7/125kHz for EU868
lora.dr = 5

print('Sending packets every minute - Interrupt to cancel loop')
print('You can send downlinks from the TTN console')
try:
    while True:
        print('Send packet')
        # Cayenne lpp random value as analog
        lora.send(bytes.fromhex('0102{:04x}'.format(randint(0, 0x7FFF))))

        while lora.nb_downlinks:
            print('Received', lora.get_downlink()['data'].hex())

        sleep(60)
except:  # noqa: E722
    pass

print('Cleaning up')
lora.close()
exit(0)



"""TTN secret keys template file.

Copy this template to secrets.py and enter you keys

Copyright 2019 Philippe Vanhaesendonck

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

SPDX-License-Identifier: Apache-2.0
"""

"""OTAA Template."""

"""Application EUI.
This EUI must be in big-endian format, so most-significant-byte
first.
For TTN issued EUIs the first bytes should be 0x70, 0xB3, 0xD5.
"""
APP_EUI = '70B3D50000000000'

"""Application key.
This key should be in big endian format (or, since it is not really a
number but a block of memory, endianness does not really apply). In
practice, a key taken from the TTN console can be copied as-is.
"""
APP_KEY = '00000000000000000000000000000000'

"""ABP Template."""

"""Device address.
The device address must be in big-endian format, so most-significant-byte
first.
For TTN issued addresses the first byte should be 0x26
"""
DEV_ADDR = '26000000'

"""Network Session Key.
The device address must be in big-endian format, so most-significant-byte
first.
"""
NWKS_KEY = '00000000000000000000000000000000'

"""App Session Key.
The device address must be in big-endian format, so most-significant-byte
first.
"""
APPS_KEY = '00000000000000000000000000000000'



"""Units tests for the CLI (V3 firmware).

Rak811 is tested separately and therefore mocked in this suite.
"""
from click.testing import CliRunner
from mock import Mock, patch, PropertyMock
from pytest import fixture
# Ignore RPi.GPIO
p = patch.dict('sys.modules', {'RPi': Mock()})
p.start()
from rak811.cli_v3 import cli  # noqa: E402
from rak811.rak811_v3 import Rak811ResponseError, Rak811TimeoutError  # noqa: E402


@fixture
def mock_rak811():
    with patch('rak811.cli_v3.Rak811', autospec=True) as p:
        yield p


@fixture
def runner():
    return CliRunner()


def test_set_config(runner, mock_rak811):
    mock_rak811.return_value.set_config.return_value = [' ']
    result = runner.invoke(cli, ['-v', 'set-config', 'lora:confirm:1'])
    mock_rak811.return_value.set_config.assert_called_once_with('lora:confirm:1')
    assert result.output == 'Configuration done\n'


def test_set_config_multi(runner, mock_rak811):
    mock_rak811.return_value.set_config.return_value = ['LoRa work mode:LoRaWAN']
    result = runner.invoke(cli, ['-v', 'set-config', 'lora:confirm:1'])
    mock_rak811.return_value.set_config.assert_called_once_with('lora:confirm:1')
    assert result.output == 'Configuration done\nLoRa work mode:LoRaWAN\n'


def test_set_config_error(runner, mock_rak811):
    mock_rak811.return_value.set_config.side_effect = Rak811ResponseError(1)
    result = runner.invoke(cli, ['-v', 'set-config', 'xxx'])
    mock_rak811.return_value.set_config.assert_called_once_with('xxx')
    assert result.output == 'RAK811 response error 1: Unsupported AT command\n'


def test_get_config(runner, mock_rak811):
    mock_rak811.return_value.get_config.return_value = ['0']
    result = runner.invoke(cli, ['-v', 'get-config', 'device:gpio:2'])
    mock_rak811.return_value.get_config.assert_called_once()
    assert result.output == '0\n'


def test_get_config_error(runner, mock_rak811):
    mock_rak811.return_value.get_config.side_effect = Rak811ResponseError(2)
    result = runner.invoke(cli, ['-v', 'get-config', 'xxx'])
    mock_rak811.return_value.get_config.assert_called_once()
    assert result.output == 'RAK811 response error 2: Invalid parameter in AT command\n'


def test_hard_reset(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'hard-reset'])
    mock_rak811.return_value.hard_reset.assert_called_once()
    assert result.output == 'Hard reset complete\n'


def test_version(runner, mock_rak811):
    p = PropertyMock(return_value='V3.0.0.14.H')
    type(mock_rak811.return_value).version = p
    result = runner.invoke(cli, ['version'])
    p.assert_called_once_with()
    assert result.output == 'V3.0.0.14.H\n'


def test_help(runner, mock_rak811):
    p = PropertyMock(return_value=['Help text'])
    type(mock_rak811.return_value).help = p
    result = runner.invoke(cli, ['help'])
    p.assert_called_once_with()
    assert result.output == 'Help text\n'


def test_run(runner, mock_rak811):
    p = PropertyMock(return_value=['Initialization OK'])
    type(mock_rak811.return_value).run = p
    runner.invoke(cli, ['run'])
    p.assert_called_once_with()


def test_send_uart(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'send-uart', 'Hello'])
    mock_rak811.return_value.send_uart.assert_called_once_with(
        data='Hello',
        index=3
    )
    assert 'Data sent.' in result.output


def test_send_uart_binary(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'send-uart', '--binary', '01020211'])
    mock_rak811.return_value.send_uart.assert_called_once_with(
        data=bytes.fromhex('01020211'),
        index=3
    )
    assert 'Data sent.' in result.output


def test_send_uart_binary_invalid(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'send-uart', '--binary', '010202xx'])
    assert result.output == 'Invalid binary data\n'


def test_send_uart_error(runner, mock_rak811):
    mock_rak811.return_value.send_uart.side_effect = Rak811ResponseError(5)
    result = runner.invoke(cli, ['-v', 'send-uart', 'Hello'])
    mock_rak811.return_value.send_uart.assert_called_once_with(
        data='Hello',
        index=3
    )
    assert result.output == 'RAK811 response error 5: Error sending through UART\n'


def test_join(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'join'])
    mock_rak811.return_value.join.assert_called_once()
    assert result.output == 'Joined!\n'


def test_join_error(runner, mock_rak811):
    mock_rak811.return_value.join.side_effect = Rak811ResponseError(99)
    result = runner.invoke(cli, ['-v', 'join'])
    mock_rak811.return_value.join.assert_called_once()
    assert result.output == 'RAK811 response error 99: LoRa join failed\n'


def test_join_timeout(runner, mock_rak811):
    mock_rak811.return_value.join.side_effect = Rak811TimeoutError(
        'Timeout while waiting for data'
    )
    result = runner.invoke(cli, ['-v', 'join'])
    mock_rak811.return_value.join.assert_called_once()
    assert result.output == 'RAK811 timeout: Timeout while waiting for data\n'


def test_send_unconfirmed(runner, mock_rak811):
    p = PropertyMock(return_value=0)
    type(mock_rak811.return_value).nb_downlinks = p
    result = runner.invoke(cli, ['-v', 'send', 'Hello'])
    mock_rak811.return_value.send.assert_called_once_with(
        data='Hello',
        port=1
    )
    assert 'Message sent.' in result.output


def test_send_confirmed(runner, mock_rak811):
    p = PropertyMock(return_value=1)
    type(mock_rak811.return_value).nb_downlinks = p
    mock_rak811.return_value.get_downlink.return_value = {
        'port': 0,
        'rssi': -34,
        'snr': 27,
        'len': 0,
        'data': '',
    }
    result = runner.invoke(cli, ['-v', 'send', '--port', '2', 'Hello'])
    mock_rak811.return_value.send.assert_called_once_with(
        data='Hello',
        port=2
    )
    assert 'Message sent.' in result.output
    assert 'Send confirmed.' in result.output


def test_send_binary(runner, mock_rak811):
    p = PropertyMock(return_value=0)
    type(mock_rak811.return_value).nb_downlinks = p
    result = runner.invoke(cli, ['-v', 'send', '--binary', '01020211'])
    mock_rak811.return_value.send.assert_called_once_with(
        data=bytes.fromhex('01020211'),
        port=1
    )
    assert 'Message sent.' in result.output
    assert 'No downlink available.' in result.output


def test_send_binary_invalid(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'send', '--binary', '010202xx'])
    assert result.output == 'Invalid binary data\n'


def test_send_error(runner, mock_rak811):
    mock_rak811.return_value.send.side_effect = Rak811ResponseError(94)
    p = PropertyMock(return_value=0)
    type(mock_rak811.return_value).nb_downlinks = p
    result = runner.invoke(cli, ['-v', 'send', '--binary', '01020211'])
    mock_rak811.return_value.send.assert_called_once_with(
        data=bytes.fromhex('01020211'),
        port=1
    )
    assert result.output == 'RAK811 response error 94: LoRa transmiting timeout\n'


def test_send_receive(runner, mock_rak811):
    p = PropertyMock(return_value=1)
    type(mock_rak811.return_value).nb_downlinks = p
    mock_rak811.return_value.get_downlink.return_value = {
        'port': 11,
        'rssi': -34,
        'snr': 27,
        'len': 4,
        'data': bytes.fromhex('65666768'),
    }
    result = runner.invoke(cli, ['-v', 'send', '--binary', '01020211'])
    mock_rak811.return_value.send.assert_called_once_with(
        data=bytes.fromhex('01020211'),
        port=1
    )
    assert 'Downlink received' in result.output
    assert 'Port: 11' in result.output
    assert 'RSSI: -34' in result.output
    assert 'SNR: 27' in result.output
    assert 'Data: 65666768' in result.output


def test_send_receive_json(runner, mock_rak811):
    p = PropertyMock(return_value=1)
    type(mock_rak811.return_value).nb_downlinks = p
    mock_rak811.return_value.get_downlink.return_value = {
        'port': 11,
        'rssi': -34,
        'snr': 27,
        'len': 4,
        'data': bytes.fromhex('65666768'),
    }
    result = runner.invoke(cli, ['-v', 'send', '--json',
                                 '--binary', '01020211'])
    mock_rak811.return_value.send.assert_called_once_with(
        data=bytes.fromhex('01020211'),
        port=1
    )
    assert ('"port": 11') in result.output
    assert ('"rssi": -34') in result.output
    assert ('"snr": 27') in result.output
    assert ('"len": 4') in result.output
    assert ('"data": "65666768"') in result.output


def test_send_p2p(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'send-p2p', 'Hello'])
    mock_rak811.return_value.send_p2p.assert_called_once_with(
        data='Hello'
    )
    assert 'Message sent.' in result.output


def test_send_p2p_binary(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'send-p2p', '--binary', '01020211'])
    mock_rak811.return_value.send_p2p.assert_called_once_with(
        data=bytes.fromhex('01020211')
    )
    assert 'Message sent.' in result.output


def test_send_p2p_binary_invalid(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'send-p2p', '--binary', '010202xx'])
    assert result.output == 'Invalid binary data\n'


def test_send_p2p_error(runner, mock_rak811):
    mock_rak811.return_value.send_p2p.side_effect = Rak811ResponseError(93)
    result = runner.invoke(cli, ['-v', 'send-p2p', 'Hello'])
    mock_rak811.return_value.send_p2p.assert_called_once_with(
        data='Hello'
    )
    assert result.output == 'RAK811 response error 93: Status is error\n'


def test_receive_p2p(runner, mock_rak811):
    p = PropertyMock(return_value=1)
    type(mock_rak811.return_value).nb_downlinks = p
    mock_rak811.return_value.get_downlink.return_value = {
        'port': 0,
        'rssi': -34,
        'snr': 27,
        'len': 4,
        'data': bytes.fromhex('65666768'),
    }
    result = runner.invoke(cli, ['-v', 'receive-p2p', '10'])
    mock_rak811.return_value.receive_p2p.assert_called_once_with(10)
    assert 'Message received' in result.output
    assert 'RSSI: -34' in result.output
    assert 'SNR: 27' in result.output
    assert 'Data: 65666768' in result.output


def test_receive_p2p_json(runner, mock_rak811):
    p = PropertyMock(return_value=1)
    type(mock_rak811.return_value).nb_downlinks = p
    mock_rak811.return_value.get_downlink.return_value = {
        'port': 0,
        'rssi': -34,
        'snr': 27,
        'len': 4,
        'data': bytes.fromhex('65666768'),
    }
    result = runner.invoke(cli, ['-v', 'receive-p2p', '--json', '10'])
    mock_rak811.return_value.receive_p2p.assert_called_once_with(10)
    assert ('"port": 0') in result.output
    assert ('"rssi": -34') in result.output
    assert ('"snr": 27') in result.output
    assert ('"len": 4') in result.output
    assert ('"data": "65666768"') in result.output



"""Units tests for the CLI.

Rak811 is tested separately and therefore mocked in this suite.
"""
from click.testing import CliRunner
from mock import Mock, patch, PropertyMock
from pytest import fixture
# Ignore RPi.GPIO
p = patch.dict('sys.modules', {'RPi': Mock()})
p.start()
from rak811.cli import cli  # noqa: E402
from rak811.rak811 import Mode, RecvEx, Reset  # noqa: E402
from rak811.rak811 import Rak811EventError, Rak811ResponseError, \
    Rak811TimeoutError  # noqa: E402


@fixture
def mock_rak811():
    with patch('rak811.cli.Rak811', autospec=True) as p:
        yield p


@fixture
def runner():
    return CliRunner()


def test_hard_reset(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'hard-reset'])
    mock_rak811.return_value.hard_reset.assert_called_once()
    assert result.output == 'Hard reset complete\n'


def test_version(runner, mock_rak811):
    p = PropertyMock(return_value='2.0.3.0')
    type(mock_rak811.return_value).version = p
    result = runner.invoke(cli, ['version'])
    p.assert_called_once_with()
    assert result.output == '2.0.3.0\n'


def test_sleep(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'sleep'])
    mock_rak811.return_value.sleep.assert_called_once()
    assert result.output == 'Sleeping\n'


def test_wake(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'wake-up'])
    mock_rak811.return_value.wake_up.assert_called_once()
    assert result.output == 'Alive!\n'


def test_reset_module(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'reset', 'module'])
    mock_rak811.return_value.reset.assert_called_once_with(Reset.Module)
    assert result.output.startswith('Module reset')


def test_reset_lora(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'reset', 'lora'])
    mock_rak811.return_value.reset.assert_called_once_with(Reset.LoRa)
    assert result.output.startswith('LoRa reset')


def test_reload(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'reload'])
    mock_rak811.return_value.reload.assert_called_once()
    assert result.output == 'Configuration reloaded.\n'


def test_mode(runner, mock_rak811):
    p = PropertyMock(return_value=Mode.LoRaWan)
    type(mock_rak811.return_value).mode = p
    result = runner.invoke(cli, ['mode'])
    p.assert_called_once_with()
    assert result.output == 'LoRaWan\n'


def test_mode_lora(runner, mock_rak811):
    p = PropertyMock()
    type(mock_rak811.return_value).mode = p
    result = runner.invoke(cli, ['-v', 'mode', 'LoRawan'])
    p.assert_called_once_with(Mode.LoRaWan)
    assert result.output == 'Mode set to LoRaWan.\n'


def test_recv_ex(runner, mock_rak811):
    p = PropertyMock(return_value=RecvEx.Enabled)
    type(mock_rak811.return_value).recv_ex = p
    result = runner.invoke(cli, ['recv-ex'])
    p.assert_called_once_with()
    assert result.output == 'Enabled\n'


def test_recv_ex_disabled(runner, mock_rak811):
    p = PropertyMock()
    type(mock_rak811.return_value).recv_ex = p
    result = runner.invoke(cli, ['-v', 'recv-ex', 'disable'])
    p.assert_called_once_with(RecvEx.Disabled)
    assert result.output == 'RSSI & SNR report on receive Disabled.\n'


def test_band(runner, mock_rak811):
    p = PropertyMock(return_value='US915')
    type(mock_rak811.return_value).band = p
    result = runner.invoke(cli, ['band'])
    p.assert_called_once_with()
    assert result.output == 'US915\n'


def test_band_eu868(runner, mock_rak811):
    p = PropertyMock()
    type(mock_rak811.return_value).band = p
    result = runner.invoke(cli, ['-v', 'band', 'EU868'])
    p.assert_called_once_with('EU868')
    assert result.output == 'LoRaWan region set to EU868.\n'


def test_set_config(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'set-config', 'dr=0', 'adr=on'])
    mock_rak811.return_value.set_config.assert_called_once_with(dr='0',
                                                                adr='on')
    assert result.output == 'LoRaWan parameters set\n'


def test_set_config_invalid(runner, mock_rak811):
    mock_rak811.return_value.set_config.side_effect = Rak811ResponseError(-1)
    result = runner.invoke(cli, ['-v', 'set-config', 'dr=0', 'adr=out'])
    mock_rak811.return_value.set_config.assert_called_once_with(dr='0',
                                                                adr='out')
    assert result.output == 'RAK811 response error -1: Invalid argument\n'


def test_set_config_nokv(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'set-config', 'dr:0'])
    assert 'dr:0 is not a valid Key=Value parameter' in result.output


def test_set_config_badkey(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'set-config', 'dx=0'])
    assert 'dx is not a valid config key' in result.output


def test_get_config(runner, mock_rak811):
    mock_rak811.return_value.get_config.return_value = 5
    result = runner.invoke(cli, ['-v', 'get-config', 'dr'])
    mock_rak811.return_value.get_config.assert_called_once()
    assert result.output == '5\n'


def test_get_config_error(runner, mock_rak811):
    mock_rak811.return_value.get_config.side_effect = Rak811ResponseError(-1)
    result = runner.invoke(cli, ['-v', 'get-config', 'nwks_key'])
    mock_rak811.return_value.get_config.assert_called_once()
    assert result.output == 'RAK811 response error -1: Invalid argument\n'


def test_join_otaa(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'join-otaa'])
    mock_rak811.return_value.join_otaa.assert_called_once()
    assert result.output == 'Joined in OTAA mode\n'


def test_join_otaa_event(runner, mock_rak811):
    mock_rak811.return_value.join_otaa.side_effect = Rak811EventError(4)
    result = runner.invoke(cli, ['-v', 'join-otaa'])
    mock_rak811.return_value.join_otaa.assert_called_once()
    assert result.output == 'RAK811 event error 4: Join failed\n'


def test_join_otaa_timeout(runner, mock_rak811):
    mock_rak811.return_value.join_otaa.side_effect = Rak811TimeoutError(
        'Timeout while waiting for event'
    )
    result = runner.invoke(cli, ['-v', 'join-otaa'])
    mock_rak811.return_value.join_otaa.assert_called_once()
    assert result.output == 'RAK811 timeout: Timeout while waiting for event\n'


def test_join_abp(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'join-abp'])
    mock_rak811.return_value.join_abp.assert_called_once()
    assert result.output == 'Joined in ABP mode\n'


def test_join_abp_response(runner, mock_rak811):
    mock_rak811.return_value.join_abp.side_effect = Rak811ResponseError(-3)
    result = runner.invoke(cli, ['-v', 'join-abp'])
    mock_rak811.return_value.join_abp.assert_called_once()
    assert result.output == 'RAK811 response error -3: ABP join error\n'


def test_signal(runner, mock_rak811):
    p = PropertyMock(return_value=(-30, 26))
    type(mock_rak811.return_value).signal = p
    result = runner.invoke(cli, ['signal'])
    p.assert_called_once_with()
    assert result.output == '-30 26\n'


def test_dr(runner, mock_rak811):
    p = PropertyMock(return_value=5)
    type(mock_rak811.return_value).dr = p
    result = runner.invoke(cli, ['dr'])
    p.assert_called_once_with()
    assert result.output == '5\n'


def test_set_dr(runner, mock_rak811):
    p = PropertyMock()
    type(mock_rak811.return_value).dr = p
    result = runner.invoke(cli, ['-v', 'dr', '5'])
    p.assert_called_once_with(5)
    assert result.output == 'Data rate set to 5.\n'


def test_link_cnt(runner, mock_rak811):
    p = PropertyMock(return_value=(15, 2))
    type(mock_rak811.return_value).link_cnt = p
    result = runner.invoke(cli, ['-v', 'link-cnt'])
    p.assert_called_once_with()
    assert result.output == 'Uplink: 15 - Downlink: 2\n'


def test_abp_info(runner, mock_rak811):
    p = PropertyMock(return_value=(
        # cSpell:disable
        '13',
        '26dddddd',
        '9annnnnnnnnnnnnnnnnnnnnnnnnnnnnn',
        '0baaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
        # cSpell:enable
    ))
    type(mock_rak811.return_value).abp_info = p
    result = runner.invoke(cli, ['abp-info'])
    p.assert_called_once_with()
    assert result.output == (
        # cSpell:disable
        '13 '
        '26dddddd '
        '9annnnnnnnnnnnnnnnnnnnnnnnnnnnnn '
        '0baaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
        '\n'
        # cSpell:enable
    )


def test_abp_info_verbose(runner, mock_rak811):
    p = PropertyMock(return_value=(
        # cSpell:disable
        '13',
        '26dddddd',
        '9annnnnnnnnnnnnnnnnnnnnnnnnnnnnn',
        '0baaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
        # cSpell:enable
    ))
    type(mock_rak811.return_value).abp_info = p
    result = runner.invoke(cli, ['-v', 'abp-info'])
    p.assert_called_once_with()
    assert result.output == (
        # cSpell:disable
        'NwkId: 13\n'
        'DevAddr: 26dddddd\n'
        'Nwkskey: 9annnnnnnnnnnnnnnnnnnnnnnnnnnnnn\n'
        'Appskey: 0baaaaaaaaaaaaaaaaaaaaaaaaaaaaaa\n'
        # cSpell:enable
    )


def test_send_unconfirmed(runner, mock_rak811):
    p = PropertyMock(return_value=0)
    type(mock_rak811.return_value).nb_downlinks = p
    result = runner.invoke(cli, ['-v', 'send', 'Hello'])
    mock_rak811.return_value.send.assert_called_once_with(
        data='Hello',
        confirm=False,
        port=1
    )
    assert 'Message sent.' in result.output


def test_send_confirmed(runner, mock_rak811):
    p = PropertyMock(return_value=0)
    type(mock_rak811.return_value).nb_downlinks = p
    result = runner.invoke(cli, ['-v', 'send', '--confirm',
                                 '--port', '2', 'Hello'])
    mock_rak811.return_value.send.assert_called_once_with(
        data='Hello',
        confirm=True,
        port=2
    )
    assert 'Message sent.' in result.output
    assert 'No downlink available.' in result.output


def test_send_binary(runner, mock_rak811):
    p = PropertyMock(return_value=0)
    type(mock_rak811.return_value).nb_downlinks = p
    result = runner.invoke(cli, ['-v', 'send', '--binary', '01020211'])
    mock_rak811.return_value.send.assert_called_once_with(
        data=bytes.fromhex('01020211'),
        confirm=False,
        port=1
    )
    assert 'Message sent.' in result.output
    assert 'No downlink available.' in result.output


def test_send_binary_invalid(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'send', '--binary', '010202xx'])
    assert result.output == 'Invalid binary data\n'


def test_send_error(runner, mock_rak811):
    mock_rak811.return_value.send.side_effect = Rak811EventError(5)
    p = PropertyMock(return_value=0)
    type(mock_rak811.return_value).nb_downlinks = p
    result = runner.invoke(cli, ['-v', 'send', '--binary', '01020211'])
    mock_rak811.return_value.send.assert_called_once_with(
        data=bytes.fromhex('01020211'),
        confirm=False,
        port=1
    )
    assert result.output == 'RAK811 event error 5: Tx timeout\n'


def test_send_receive_recv_tx(runner, mock_rak811):
    p = PropertyMock(return_value=1)
    type(mock_rak811.return_value).nb_downlinks = p
    mock_rak811.return_value.get_downlink.return_value = {
        'port': 11,
        'rssi': -34,
        'snr': 27,
        'len': 4,
        'data': bytes.fromhex('65666768'),
    }
    result = runner.invoke(cli, ['-v', 'send', '--binary', '01020211'])
    mock_rak811.return_value.send.assert_called_once_with(
        data=bytes.fromhex('01020211'),
        confirm=False,
        port=1
    )
    assert 'Downlink received' in result.output
    assert 'Port: 11' in result.output
    assert 'RSSI: -34' in result.output
    assert 'SNR: 27' in result.output
    assert 'Data: 65666768' in result.output


def test_send_receive(runner, mock_rak811):
    p = PropertyMock(return_value=1)
    type(mock_rak811.return_value).nb_downlinks = p
    mock_rak811.return_value.get_downlink.return_value = {
        'port': 11,
        'rssi': 0,
        'snr': 0,
        'len': 4,
        'data': bytes.fromhex('65666768'),
    }
    result = runner.invoke(cli, ['-v', 'send', '--binary', '01020211'])
    mock_rak811.return_value.send.assert_called_once_with(
        data=bytes.fromhex('01020211'),
        confirm=False,
        port=1
    )
    assert 'Downlink received' in result.output
    assert 'Port: 11' in result.output
    assert 'RSSI' not in result.output
    assert 'SNR' not in result.output
    assert 'Data: 65666768' in result.output


def test_send_receive_json(runner, mock_rak811):
    p = PropertyMock(return_value=1)
    type(mock_rak811.return_value).nb_downlinks = p
    mock_rak811.return_value.get_downlink.return_value = {
        'port': 11,
        'rssi': -34,
        'snr': 27,
        'len': 4,
        'data': bytes.fromhex('65666768'),
    }
    result = runner.invoke(cli, ['-v', 'send', '--json',
                                 '--binary', '01020211'])
    mock_rak811.return_value.send.assert_called_once_with(
        data=bytes.fromhex('01020211'),
        confirm=False,
        port=1
    )
    assert ('"port": 11') in result.output
    assert ('"rssi": -34') in result.output
    assert ('"snr": 27') in result.output
    assert ('"len": 4') in result.output
    assert ('"data": "65666768"') in result.output


def test_set_rf_config_invalid_parameter(runner, mock_rak811):
    result = runner.invoke(cli, ['rf-config', 'tx=8'])
    assert (
        "Error: Invalid value for 'KEY=VALUE...': "
        'tx is not a valid config key'
    ) in result.output


def test_set_rf_config_invalid_kv(runner, mock_rak811):
    result = runner.invoke(cli, ['rf-config', 'sf:8'])
    assert (
        "Error: Invalid value for 'KEY=VALUE...': "
        'sf:8 is not a valid Key=Value parameter'
    ) in result.output


def test_set_rf_config_invalid_range(runner, mock_rak811):
    result = runner.invoke(cli, ['rf-config', 'sf=1'])
    assert (
        "Error: Invalid value for 'KEY=VALUE...': "
        '1 is not in the range 6<=x<=12.\n'
    ) in result.output


def test_set_rf_config_one(runner, mock_rak811):
    p = PropertyMock()
    type(mock_rak811.return_value).rf_config = p
    result = runner.invoke(cli, ['-v', 'rf-config', 'sf=8'])
    p.assert_called_once_with({
        'sf': 8
    })
    assert result.output == 'rf_config set: sf=8\n'


def test_set_rf_config_all(runner, mock_rak811):
    p = PropertyMock()
    type(mock_rak811.return_value).rf_config = p
    result = runner.invoke(cli, [
        '-v', 'rf-config',
        'freq=868.200',
        'sf=8',
        'bw=1',
        'cr=2',
        'prlen=16',
        'pwr=8'
    ])
    p.assert_called_once_with({
        'freq': 868.200,
        'sf': 8,
        'bw': 1,
        'cr': 2,
        'prlen': 16,
        'pwr': 8
    })
    assert('freq=868.2') in result.output
    assert('sf=8') in result.output
    assert('bw=1') in result.output
    assert('cr=2') in result.output
    assert('prlen=16') in result.output
    assert('pwr=8') in result.output


def test_get_rf_config(runner, mock_rak811):
    p = PropertyMock(return_value={
        'freq': 868.200,
        'sf': 8,
        'bw': 1,
        'cr': 2,
        'prlen': 16,
        'pwr': 8
    })
    type(mock_rak811.return_value).rf_config = p
    result = runner.invoke(cli, ['rf-config'])
    p.assert_called_once_with()
    assert result.output == '868.2 8 1 2 16 8\n'


def test_get_rf_config_verbose(runner, mock_rak811):
    p = PropertyMock(return_value={
        'freq': 868.200,
        'sf': 8,
        'bw': 1,
        'cr': 2,
        'prlen': 16,
        'pwr': 8
    })
    type(mock_rak811.return_value).rf_config = p
    result = runner.invoke(cli, ['-v', 'rf-config'])
    p.assert_called_once_with()
    assert result.output == (
        'Frequency: 868.2\n'
        'SF: 8\n'
        'BW: 1\n'
        'CR: 2\n'
        'PrLen: 16\n'
        'Power: 8\n'
    )


def test_txc(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'txc', 'Hello'])
    mock_rak811.return_value.txc.assert_called_once_with(
        data='Hello',
        cnt=1,
        interval=60
    )
    assert result.output == 'Message sent.\n'


def test_txc_binary(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'txc', '--binary', '01020211'])
    mock_rak811.return_value.txc.assert_called_once_with(
        data=bytes.fromhex('01020211'),
        cnt=1,
        interval=60
    )
    assert result.output == 'Message sent.\n'


def test_txc_binary_invalid(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'txc', '--binary', '010202xx'])
    assert result.output == 'Invalid binary data\n'


def test_txc_error(runner, mock_rak811):
    mock_rak811.return_value.txc.side_effect = Rak811EventError(5)
    result = runner.invoke(cli, ['-v', 'txc', 'Hello'])
    mock_rak811.return_value.txc.assert_called_once_with(
        data='Hello',
        cnt=1,
        interval=60
    )
    assert result.output == 'RAK811 event error 5: Tx timeout\n'


def test_rxc(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'rxc'])
    mock_rak811.return_value.rxc.assert_called_once()
    assert result.output == 'Module set in receive mode.\n'


def test_tx_stop(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'tx-stop'])
    mock_rak811.return_value.tx_stop.assert_called_once()
    assert result.output == 'LoraP2P TX stopped.\n'


def test_rx_stop(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'rx-stop'])
    mock_rak811.return_value.rx_stop.assert_called_once()
    assert result.output == 'LoraP2P RX stopped.\n'


def test_rx_get_no_message(runner, mock_rak811):
    p = PropertyMock(return_value=0)
    type(mock_rak811.return_value).nb_downlinks = p
    result = runner.invoke(cli, ['-v', 'rx-get', '0'])
    mock_rak811.return_value.rx_get.assert_called_once_with(0)
    assert result.output == 'No message available.\n'


def test_rx_get_message(runner, mock_rak811):
    p = PropertyMock(return_value=1)
    type(mock_rak811.return_value).nb_downlinks = p
    mock_rak811.return_value.get_downlink.return_value = {
        'port': 0,
        'rssi': -34,
        'snr': 27,
        'len': 4,
        'data': bytes.fromhex('65666768'),
    }
    result = runner.invoke(cli, ['rx-get', '0'])
    mock_rak811.return_value.rx_get.assert_called_once_with(0)
    assert result.output == '65666768\n'


def test_rx_get_message_verbose(runner, mock_rak811):
    p = PropertyMock(return_value=1)
    type(mock_rak811.return_value).nb_downlinks = p
    mock_rak811.return_value.get_downlink.return_value = {
        'port': 0,
        'rssi': -34,
        'snr': 27,
        'len': 4,
        'data': bytes.fromhex('65666768'),
    }
    result = runner.invoke(cli, ['-v', 'rx-get', '0'])
    mock_rak811.return_value.rx_get.assert_called_once_with(0)
    assert 'RSSI: -34' in result.output
    assert 'SNR: 27' in result.output
    assert 'Data: 65666768' in result.output


def test_rx_get_message_json(runner, mock_rak811):
    p = PropertyMock(return_value=1)
    type(mock_rak811.return_value).nb_downlinks = p
    mock_rak811.return_value.get_downlink.return_value = {
        'port': 0,
        'rssi': 0,
        'snr': 0,
        'len': 4,
        'data': bytes.fromhex('65666768'),
    }
    result = runner.invoke(cli, ['rx-get', '--json', '0'])
    mock_rak811.return_value.rx_get.assert_called_once_with(0)
    assert '"port": 0' in result.output
    assert '"data": "65666768"' in result.output


def test_radio_status(runner, mock_rak811):
    p = PropertyMock(return_value=(8, 0, 1, 0, 0, -48, 28))
    type(mock_rak811.return_value).radio_status = p
    result = runner.invoke(cli, ['radio-status'])
    p.assert_called_once_with()
    assert result.output == (
        '8 0 1 0 0 -48 28\n'
    )


def test_radio_status_verbose(runner, mock_rak811):
    p = PropertyMock(return_value=(8, 0, 1, 0, 0, -48, 28))
    type(mock_rak811.return_value).radio_status = p
    result = runner.invoke(cli, ['-v', 'radio-status'])
    p.assert_called_once_with()
    assert result.output == (
        'TxSuccessCnt: 8\n'
        'TxErrCnt: 0\n'
        'RxSuccessCnt: 1\n'
        'RxTimeOutCnt: 0\n'
        'RxErrCnt: 0\n'
        'RSSI: -48\n'
        'SNR: 28\n'
    )


def test_clear_radio_status(runner, mock_rak811):
    result = runner.invoke(cli, ['-v', 'clear-radio-status'])
    mock_rak811.return_value.clear_radio_status.assert_called_once()
    assert result.output == 'Radio statistics cleared.\n'



#!/usr/bin/env python3
"""RAK811 OTAA Class C demo.

Minimalistic OTAA demo with Class C device (v3.x firmware).

The module will wait on downlink message, run a task and send an the same data
back as uplink

The module uses receive_p2p to get out-of-band downlinks.

The device must be configured as a Class C device in your LoRaWan Application
Server (TTN/TTS).

Note that delivery of downlinks is not guaranteed:
    - Device might not receive the signal
    - The device is not listening when in send mode

Copyright 2022 Philippe Vanhaesendonck

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

SPDX-License-Identifier: Apache-2.0
"""
import logging
from random import randint
from sys import exit
from time import sleep
from timeit import default_timer as timer
from traceback import print_exc

from rak811.rak811_v3 import Rak811
from ttn_secrets import APP_EUI, APP_KEY

# Set level to logging.DEBUG for a verbose output
logging.basicConfig(level=logging.INFO)

lora = Rak811()

# Most of the setup should happen only once...
print("Setup")
# Ensure we are in LoRaWan mode / Class C
lora.set_config("lora:work_mode:0")
lora.set_config("lora:class:2")
# Select OTAA
lora.set_config("lora:join_mode:0")
# Select region
lora.set_config("lora:region:EU868")
# Set keys
lora.set_config(f"lora:app_eui:{APP_EUI}")
lora.set_config(f"lora:app_key:{APP_KEY}")
# Set data rate
# Note that DR is different from SF and depends on the region
# See: https://docs.exploratory.engineering/lora/dr_sf/
# Set Data Rate to 5 which is SF7/125kHz for EU868
lora.set_config("lora:dr:5")

# Print config
for line in lora.get_config("lora:status"):
    print(f"    {line}")

print("Joining")
start_time = timer()
lora.join()
print("Joined in {:.2f} secs".format(timer() - start_time))

print("Sending initial Hello packet")
start_time = timer()
lora.send("Hello")
print("Packet sent in {:.2f} secs".format(timer() - start_time))
print("Entering wait loop")
print("You can send downlinks from the TTN console")
try:
    while True:
        print("Waiting for downlinks...")
        lora.receive_p2p(60)
        while lora.nb_downlinks:
            data = lora.get_downlink()["data"]
            if data != b"":
                print("Downlink received", data.hex())
                # simulate some processing time
                sleep(randint(5, 10))
                print("Sending back results")
                start_time = timer()
                lora.send(data)
                print("Packet sent in {:.2f} secs".format(timer() - start_time))

except KeyboardInterrupt:
    print()
except Exception:
    print_exc()

print("Cleaning up")
lora.close()
exit(0)



"""Units tests for the Rak811 class (V3 firmware).

Rak811Serial is tested separately and therefore mocked in this suite.

RPi.GPIO is completely ignored as not available on all platforms, and its use
by the Rak811 class very limited.
"""
from mock import Mock, patch
from pytest import fixture, raises
# Ignore RPi.GPIO
p = patch.dict('sys.modules', {'RPi': Mock()})
p.start()
from rak811.rak811_v3 import EVENT_TIMEOUT, RESPONSE_TIMEOUT  # noqa: E402
from rak811.rak811_v3 import Rak811, Rak811ResponseError  # noqa: E402
from rak811.rak811_v3 import Rak811Serial  # noqa: E402
from rak811.serial import Rak811TimeoutError  # noqa: E402


@fixture
def lora():
    """Instantiate Rak811 class though fixture."""
    with patch('rak811.rak811_v3.Rak811Serial', autospec=True):
        rak811 = Rak811()
        yield rak811
        rak811.close()


@patch('rak811.rak811_v3.Rak811Serial', autospec=True)
def test_instantiate_default(mock_serial):
    """Test that Rak811 can be instantiated.

    Check for basic initialisation and teardown of the RackSerial.
    """
    lora = Rak811()

    assert isinstance(lora._serial, Rak811Serial)
    mock_serial.assert_called_once_with(
        keep_untagged=True,
        read_buffer_timeout=RESPONSE_TIMEOUT
    )
    lora.close()
    mock_serial.return_value.close.assert_called_once()


@patch('rak811.rak811_v3.Rak811Serial', autospec=True)
def test_instantiate_params(mock_serial):
    """Test that Rak811 passes parameters passed to RackSerial."""
    port = '/dev/ttyAMA0'
    timeout = 5
    lora = Rak811(port=port, timeout=timeout)
    mock_serial.assert_called_once_with(
        keep_untagged=True,
        read_buffer_timeout=RESPONSE_TIMEOUT,
        port=port,
        timeout=timeout
    )
    lora.close()


@patch('rak811.rak811_v3.GPIO')
@patch('rak811.rak811_v3.sleep')
def test_hard_reset(mock_sleep, mock_gpio, lora):
    """Test hard_reset().

    Test is a bit pointless, at least ensures it runs...
    """
    lora.hard_reset()
    mock_gpio.setup.assert_called_once()
    mock_gpio.cleanup.assert_not_called()


def test_int(lora):
    """Test _int function."""
    assert lora._int(1) == 1
    assert lora._int('1') == 1
    assert lora._int('Hello') == 'Hello'


def test_send_string(lora):
    """Test _send_string (passthrough to RackSerial)."""
    lora._send_string('Hello')
    lora._serial.send_string.assert_called_with('Hello')


def test_send_command(lora):
    """Test _send_command."""
    # Successful command
    lora._serial.receive.return_value = 'OK V3.0.0.14.H'
    assert lora._send_command('version') == 'V3.0.0.14.H'
    lora._serial.send_command.assert_called_with('version')

    # Error
    lora._serial.receive.return_value = 'ERROR: 2'
    with raises(Rak811ResponseError,
                match='2'):
        lora._send_command('set_config=lora:work_mode:5')

    # Unexpected data  response queue
    lora._serial.receive.side_effect = [
        'Unexpected',
        'OK V3.0.0.14.H',
    ]
    assert lora._send_command('version') == 'V3.0.0.14.H'

    # Multi-line response with "Initialization OK "
    lora._serial.receive.side_effect = [
        'RAK811 Version:3.0.0.14.H',
        'LoRa work mode:LoRaWAN, join_mode:OTAA, MulticastEnable: false, Class: A',
        'Initialization OK '
    ]
    assert lora._send_command('set_config=device:restart') == ' '


def test_send_command_list(lora):
    """Test _send_command_list."""
    # OK message
    lora._serial.receive.return_value = [
        'OK Device AT commands:',
        '  at+version',
        '  at+send=lorap2p:XXX'
    ]
    assert lora._send_command_list('help') == [
        'Device AT commands:',
        '  at+version',
        '  at+send=lorap2p:XXX'
    ]

    # "Initialization OK "
    lora._serial.receive.return_value = [
        'RAK811 Version:3.0.0.14.H',
        'LoRa work mode:LoRaWAN, join_mode:OTAA, MulticastEnable: false, Class: A',
        'Initialization OK '
    ]
    assert lora._send_command_list('set_config=device:restart') == [
        'RAK811 Version:3.0.0.14.H',
        'LoRa work mode:LoRaWAN, join_mode:OTAA, MulticastEnable: false, Class: A',
        ' '
    ]

    # Error
    lora._serial.receive.return_value = ['ERROR: 1']
    with raises(Rak811ResponseError,
                match='1'):
        lora._send_command_list('set_config=device:status')


def test_get_events(lora):
    """Test _get_events."""
    # Successful command
    lora._serial.receive.return_value = [
        'at+recv=0,-68,7,0',
        'at+recv=1,-65,6,2:4865',
    ]
    events = lora._get_events()
    assert events.pop() == '1,-65,6,2:4865'
    assert events.pop() == '0,-68,7,0'


"""AT command API.

For the sake of simplicity we mock Rak811._send_command,
Rak811._send_command_list and Rack811._get_events, as these have already been
tested.
"""


@patch.object(Rak811, '_send_command_list', return_value=[' '])
def test_set_config_simple(mock_send, lora):
    """Test set_config command (simple case)."""
    assert lora.set_config('lora:region:EU868') == [' ']
    mock_send.assert_called_once_with('set_config=lora:region:EU868')


@patch.object(Rak811, '_send_command_list', return_value=[' ', ' <BOOT MODE>'])
def test_set_config_boot(mock_send, lora):
    """Test set_config command (Boot case)."""
    assert lora.set_config('device:boot') == [' ', ' <BOOT MODE>']
    mock_send.assert_called_once_with('set_config=device:boot')


@patch.object(Rak811, '_send_command_list', return_value=['LoRa work mode:LoRaWAN', ' '])
def test_set_config_init(mock_send, lora):
    """Test set_config command (Initialization OK case)."""
    assert lora.set_config('lora:work_mode:0') == ['LoRa work mode:LoRaWAN', ' ']
    mock_send.assert_called_once_with('set_config=lora:work_mode:0')


@patch.object(Rak811, '_send_command_list', return_value=['1'])
def test_get_config_simple(mock_send, lora):
    """Test get_config command (simple case)."""
    assert lora.get_config('device:gpio:2') == ['1']
    mock_send.assert_called_once_with('get_config=device:gpio:2')


@patch.object(Rak811, '_send_command_list', return_value=['Work Mode: LoRaWAN', 'DownLinkCounter: 0'])
def test_get_config_multi(mock_send, lora):
    """Test get_config command (Multiple lines)."""
    assert lora.get_config('lora:status') == ['Work Mode: LoRaWAN', 'DownLinkCounter: 0']
    mock_send.assert_called_once_with('get_config=lora:status')


@patch.object(Rak811, '_send_command', return_value='V3.0.0.14.H')
def test_version(mock_send, lora):
    """Test version command."""
    assert lora.version == 'V3.0.0.14.H'
    mock_send.assert_called_once_with('version')


@patch.object(Rak811, '_send_command_list', return_value=['Device AT commands:', '  at+send=lorap2p:XXX'])
def test_help(mock_send, lora):
    """Test help command."""
    assert lora.help == ['Device AT commands:', '  at+send=lorap2p:XXX']
    mock_send.assert_called_once_with('help')


@patch.object(Rak811, '_send_command')
def test_run(mock_send, lora):
    """Test run command."""
    lora.run()
    mock_send.assert_called_once_with('run')


@patch.object(Rak811, '_send_command')
def test_send_uart(mock_send, lora):
    """Test UART send."""
    lora.send_uart('Hello')
    mock_send.assert_called_once_with('send=uart:3:48656c6c6f')


@patch.object(Rak811, '_send_command')
def test_join(mock_send, lora):
    """Test join command."""
    lora.join()
    mock_send.assert_called_once_with('join', timeout=EVENT_TIMEOUT)


@patch.object(Rak811, '_get_events')
@patch.object(Rak811, '_send_command')
def test_send_unconfirmed(mock_send, mock_events, lora):
    """Test send, unconfirmed."""
    mock_events.side_effect = Rak811TimeoutError()
    lora.send('Hello')
    mock_send.assert_called_once_with('send=lora:1:48656c6c6f', timeout=300)
    mock_events.assert_called_once()
    assert lora.nb_downlinks == 0


@patch.object(Rak811, '_get_events', return_value=['0,-68,7,0'])
@patch.object(Rak811, '_send_command')
def test_send_confirmed(mock_send, mock_events, lora):
    """Test send, unconfirmed."""
    lora.send('Hello')
    mock_send.assert_called_once_with('send=lora:1:48656c6c6f', timeout=300)
    mock_events.assert_called_once()
    assert lora.nb_downlinks == 1
    assert lora.get_downlink() == {
        'port': 0,
        'rssi': -68,
        'snr': 7,
        'len': 0,
        'data': b'',
    }


@patch.object(Rak811, '_get_events', return_value=['1,-67,8,3:313233'])
@patch.object(Rak811, '_send_command')
def test_send_downlink(mock_send, mock_events, lora):
    """Test send, unconfirmed."""
    lora.send('Hello')
    mock_send.assert_called_once_with('send=lora:1:48656c6c6f', timeout=300)
    mock_events.assert_called_once()
    assert lora.nb_downlinks == 1
    assert lora.get_downlink() == {
        'port': 1,
        'rssi': -67,
        'snr': 8,
        'len': 3,
        'data': bytes.fromhex('313233'),
    }


@patch.object(Rak811, '_send_command')
def test_send_p2p(mock_send, lora):
    """Test P2P send."""
    lora.send_p2p('Hello')
    mock_send.assert_called_once_with('send=lorap2p:48656c6c6f')


@patch.object(Rak811, '_get_events', return_value=['-67,8,3:313233'])
def test_receive_p2p(mock_events, lora):
    """Test receive P2P."""
    lora.receive_p2p(timeout=10)
    mock_events.assert_called_once_with(10)
    assert lora.nb_downlinks == 1
    assert lora.get_downlink() == {
        'port': 0,
        'rssi': -67,
        'snr': 8,
        'len': 3,
        'data': bytes.fromhex('313233'),
    }



"""RAK811 serial communication layer.

Copyright 2019, 2021 Philippe Vanhaesendonck

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

SPDX-License-Identifier: Apache-2.0
"""
from logging import getLogger
from re import match
from threading import Condition, Event, Thread
from time import sleep
from typing import List, Union

from rak811.exception import Rak811Error
from serial import Serial

# Default instance parameters. Can be overridden  at creation
# Serial port configuration
PORT = '/dev/serial0'
BAUDRATE = 115200
# Timeout for the reader thread. Any value will do, the only impact is the time
# needed to stop the thread when the instance is destroyed...
TIMEOUT = 2
# Timeout for response and events
# The RAK811 typically respond in less than 1.5 seconds
READ_BUFFER_TIMEOUT = 5

# Constants
EOL = '\r\n'


logger = getLogger(__name__)


class Rak811TimeoutError(Rak811Error):
    """Read timeout exception."""

    pass


class Rak811Serial(object):
    """Handles serial communication between the RPi and the RAK811 module."""

    def __init__(self,
                 port=PORT,
                 baudrate=BAUDRATE,
                 timeout=TIMEOUT,
                 read_buffer_timeout=READ_BUFFER_TIMEOUT,
                 keep_untagged=False,
                 **kwargs):
        """Initialise class.

        The serial port is immediately opened and flushed.
        All parameters are optional and passed to Serial.
        """
        self._read_buffer_timeout = read_buffer_timeout
        self._serial = Serial(port=port,
                              baudrate=baudrate,
                              timeout=timeout,
                              **kwargs)
        self._serial.reset_input_buffer()

        self._keep_untagged = keep_untagged

        # Mutex
        self._cv_serial = Condition()
        self._read_buffer = []

        # Read thread
        self._read_done = Event()
        self._read_thread = Thread(target=self._read_loop, daemon=True)
        self._read_thread.start()

        self._alive = True
        logger.debug('Serial initialized')

    def close(self):
        """Release resources."""
        if self._alive:
            self._read_done.set()
            self._read_thread.join()
            self._serial.close()
            self._alive = False

    def _read_loop(self):
        """Read thread.

        Continuously read serial. When data is available we want to read all of
        it and notify once:
            - We need to drain the input after a response. If we notify()
            too early the module will miss next command
            - We want to catch all events at the same time
        """
        while not self._read_done.is_set():
            line = self._serial.readline()
            if line != b'':
                # Not a timeout; process data stream
                with self._cv_serial:
                    while True:
                        try:
                            line = line.decode('ascii').rstrip(EOL)
                        except UnicodeDecodeError:
                            # Wrong speed or port not configured properly
                            line = '?'
                        if match(r'^(OK|ERROR|at+)', line):
                            logger.debug(f'Received: >{line}<')
                            self._read_buffer.append(line)
                        elif self._keep_untagged:
                            logger.debug(f'Received untagged: >{line}<')
                            self._read_buffer.append(line)
                        else:
                            logger.debug(f'Ignoring untagged: >{line}<')
                        sleep(0.1)
                        if self._serial.in_waiting > 0:
                            line = self._serial.readline()
                        else:
                            break
                    if len(self._read_buffer) > 0:
                        self._cv_serial.notify()

    def receive(self, single: bool = True, timeout: int = None) -> Union[str, List[str]]:
        """Receive data from module.

        This is a blocking call: it will data or raise Rak811TimeoutError if
        nothing is received in time.

        Args:
            single (optional): Return single line of data when true, otherwise
                all available lines are returned. Defaults to True.
            timeout (optional): Time to wait for. Defaults to None.

        Raises:
            Rak811TimeoutError: No data received in time.

        Returns:
            Single line of data or list of lines.
        """
        if timeout is None:
            timeout = self._read_buffer_timeout

        with self._cv_serial:
            while len(self._read_buffer) == 0:
                success = self._cv_serial.wait(timeout)
                if not success:
                    raise Rak811TimeoutError('Timeout while waiting for data')
            if single:
                response = self._read_buffer.pop(0)
            else:
                response = self._read_buffer
                self._read_buffer = []
        return response

    def send_string(self, string):
        """Send string to the module."""
        logger.debug(f"Sending: >{string.encode('unicode_escape').decode('utf-8')}<")
        self._serial.write((bytes)(string, 'utf-8'))

    def send_command(self, command):
        """Send AT command to the module."""
        self.send_string('at+{0}\r\n'.format(command))



"""Interface with the RAK811 module (Firmware V3.0).

Copyright 2021, 2022 Philippe Vanhaesendonck

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

SPDX-License-Identifier: Apache-2.0
"""
from binascii import hexlify
from enum import IntEnum
from logging import getLogger
from re import match
from time import sleep
from typing import List, Union

from RPi import GPIO

from .exception import Rak811Error
from .serial import Rak811Serial, Rak811TimeoutError

RESET_BCM_PORT = 17
RESET_DELAY = 0.01
RESET_POST = 2
RESPONSE_OK = 'OK '
RESPONSE_INIT_OK = 'Initialization OK'
RESPONSE_ERROR = 'ERROR:'
RESPONSE_EVENT = 'at+recv='

# Timeout for response and events
# The RAK811 typically respond in less than 1.5 seconds
RESPONSE_TIMEOUT = 5
# Event wait time strongly depends on duty cycle, when sending often at high SF
# the module will wait to respect the duty cycle.
# In normal operation, 5 minutes should be more than enough.
EVENT_TIMEOUT = 5 * 60

logger = getLogger(__name__)


# RAK811 error codes and associated messages
class ErrorCode(IntEnum):
    """AT commands error codes."""

    ERR_001 = 1
    ERR_002 = 2
    ERR_003 = 3
    ERR_004 = 4
    ERR_005 = 5
    ERR_041 = 41
    ERR_080 = 80
    ERR_081 = 81
    ERR_082 = 82
    ERR_083 = 83
    ERR_084 = 84
    ERR_085 = 85
    ERR_086 = 86
    ERR_087 = 87
    ERR_088 = 88
    ERR_089 = 89
    ERR_090 = 90
    ERR_091 = 91
    ERR_092 = 92
    ERR_093 = 93
    ERR_094 = 94
    ERR_095 = 95
    ERR_096 = 96
    ERR_097 = 97
    ERR_098 = 98
    ERR_099 = 99
    ERR_100 = 100
    ERR_101 = 101
    ERR_102 = 102
    ERR_103 = 102
    ERR_104 = 104
    ERR_INVALID_EVENT = 998
    ERR_UNKNOWN = 999


ERROR_MESSAGE = {
    ErrorCode.ERR_001: 'Unsupported AT command',
    ErrorCode.ERR_002: 'Invalid parameter in AT command',
    ErrorCode.ERR_003: 'Error when reading or writing flash',
    ErrorCode.ERR_004: 'Error reading or writing IIC',
    ErrorCode.ERR_005: 'Error sending through UART',
    ErrorCode.ERR_041: 'BLE in invalid state',
    ErrorCode.ERR_080: 'LoRa busy',
    ErrorCode.ERR_081: 'LoRa service unknown',
    ErrorCode.ERR_082: 'Invalid LoRa parameters',
    ErrorCode.ERR_083: 'Invalid LoRa frequency',
    ErrorCode.ERR_084: 'Invalid LoRa datarate (DR)',
    ErrorCode.ERR_085: 'Invalid LoRa frequency and datarate',
    ErrorCode.ERR_086: 'Device has not joined LoRa network',
    ErrorCode.ERR_087: 'Packet length too long',
    ErrorCode.ERR_088: 'Service closed by server',
    ErrorCode.ERR_089: 'Unsupported region.',
    ErrorCode.ERR_090: 'Restricted duty cycle',
    ErrorCode.ERR_091: 'No valid channel can be found.',
    ErrorCode.ERR_092: 'No free channel found',
    ErrorCode.ERR_093: 'Status is error',
    ErrorCode.ERR_094: 'LoRa transmiting timeout',
    ErrorCode.ERR_095: 'LoRa RX1 timeout',
    ErrorCode.ERR_096: 'LoRa RX2 timeout',
    ErrorCode.ERR_097: 'Error receiving RX1',
    ErrorCode.ERR_098: 'Error  receiving RX2',
    ErrorCode.ERR_099: 'LoRa join failed',
    ErrorCode.ERR_100: 'Repeated downlink',
    ErrorCode.ERR_101: 'Payload size error with transmit DR',
    ErrorCode.ERR_102: 'Too many downlink frames lost',
    ErrorCode.ERR_103: 'Address fail',
    ErrorCode.ERR_104: 'Error verifying MIC',
    ErrorCode.ERR_INVALID_EVENT: 'Invalid event received',
    ErrorCode.ERR_UNKNOWN: 'Unknown error',
}


class Rak811ResponseError(Rak811Error):
    """Exception raised by response from the module.

    Attributes:
        errno -- as returned by the module
        strerror -- textual representation

    """

    def __init__(self, code):
        """Just assign return codes."""
        try:
            self.errno = int(code)
        except ValueError:
            self.errno = code

        if self.errno in ERROR_MESSAGE:
            self.strerror = ERROR_MESSAGE[self.errno]
        else:
            self.strerror = ERROR_MESSAGE[ErrorCode.ERR_UNKNOWN]
        super().__init__(('[Errno {}] {}').format(self.errno, self.strerror))


class Rak811(object):
    """Main class."""

    def __init__(self, **kwargs):
        """Initialise class.

        The serial port is immediately opened and flushed.

        Args:
            response_timeout (optional): override default response timeout.
                Defaults to None.
            event_timeout (optional): override default event timeout.
                Defaults to None.
            Remainder parameters are passed to RackSerial.
        """
        self._response_timeout = kwargs.pop('response_timeout', RESPONSE_TIMEOUT)
        self._event_timeout = kwargs.pop('event_timeout', EVENT_TIMEOUT)
        self._serial = Rak811Serial(
            keep_untagged=True,
            read_buffer_timeout=self._response_timeout,
            **kwargs
        )
        self._downlink = []

    def close(self) -> None:
        """Terminates session.

        Terminates read thread and close serial port.
        """
        self._serial.close()

    def hard_reset(self) -> None:
        """Hard reset of the RAK811 module.

        Hard reset should not be required in normal operation. It needs to be
        issued once after host boot, or module restart.
        Note that we do not cleanup() as the reset port should stay high (it is
        configured that way at boot time).

        Note: left for historical reasons, it does not appear to be effective
        with the V3 firmware
        """
        GPIO.setwarnings(False)
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(RESET_BCM_PORT, GPIO.OUT)
        GPIO.output(RESET_BCM_PORT, GPIO.LOW)
        sleep(RESET_DELAY)
        GPIO.output(RESET_BCM_PORT, GPIO.HIGH)
        sleep(RESET_POST)

    """ Private methods."""

    def _int(self, i: str) -> Union[int, str]:
        """Attempt int conversion.

        Args:
            i: Integer string.

        Returns:
            Integer conversion if possible, else the original string.
        """
        try:
            i = int(i)
        except ValueError:
            pass
        return i

    def _send_string(self, string: str) -> None:
        """Send string to the RAK811 module.

        Args:
            string: String to send.
        """
        self._serial.send_string(string)

    def _send_command(self, command: str, timeout: float = None) -> str:
        """Send AT command to the RAK811 module and return the response.

        This method will wait until the module answer or timeout is reached.

        Args:
            command: Command to send
            timeout (optional): Time to wait before raising a timeout.
                Defaults to None.

        Raises:
            Rak811ResponseError: Module answered with an error code.
            Rack811TimeoutError: No answer received and timeout is reached.

        Returns:
            Module answer.
        """
        if timeout is None:
            timeout = self._response_timeout

        # Process possible pending events
        try:
            self._process_events(timeout=0.01)
            logger.debug('Pending event queue flushed')
        except (Rak811ResponseError, Rak811TimeoutError):
            pass

        self._serial.send_command(command)
        response = self._serial.receive(timeout=timeout)

        # Ignore anything but OK/ERROR messages
        while not (response.startswith(RESPONSE_OK)
                   or response.startswith(RESPONSE_INIT_OK)
                   or response.startswith(RESPONSE_ERROR)):
            response = self._serial.receive(timeout=timeout)

        if response.startswith(RESPONSE_ERROR):
            raise Rak811ResponseError(response[len(RESPONSE_ERROR):])
        elif response.startswith(RESPONSE_INIT_OK):
            return response[len(RESPONSE_INIT_OK):]
        else:
            return response[len(RESPONSE_OK):]

    def _send_command_list(self, command: str) -> List[str]:
        """Send AT command to the RAK811 module and return the responses.

        Similar to _send_command, but returns the complete read buffer. It is
        used for commands returning several lines.

        This method will wait until the module answer or timeout is reached.

        Args:
            command: Command to send

        Raises:
            Rak811ResponseError: Module answered with an error code.
            Rack811TimeoutError: No answer received and timeout is reached.

        Returns:
            List of module answers.
        """
        self._serial.send_command(command)

        prelude = []
        response = []
        while True:
            if not response:
                response = self._serial.receive(single=False)
            if response[0].startswith(RESPONSE_OK):
                response[0] = response[0][len(RESPONSE_OK):]
                break
            elif response[0].startswith(RESPONSE_INIT_OK):
                response[0] = response[0][len(RESPONSE_INIT_OK):]
                break
            elif response[0].startswith(RESPONSE_ERROR):
                raise Rak811ResponseError(response[0][len(RESPONSE_ERROR):])
            else:
                # Ignore anything until we get an OK/ERROR message
                prelude.append(response.pop(0))

        return prelude + response

    def _get_events(self, timeout: float = None) -> List[str]:
        """Get list of events from the RAK811 module.

        Args:
            timeout (optional): Time to wait before raising a timeout.
                Defaults to None.

        Raises:
            Rack811TimeoutError: No answer received and timeout is reached.

        Returns:
            Event list
        """
        if timeout is None:
            timeout = self._event_timeout

        return [i[len(RESPONSE_EVENT):] if i.startswith(RESPONSE_EVENT) else i for i in
                self._serial.receive(single=False, timeout=timeout)]

    def _add_downlink(self, port: str, rssi: str, snr: str, length: str, data: str) -> None:
        """Add message to the downlink list.

        Args:
            port: Message port.
            rssi: Message RSSI.
            snr: Message SNR.
            length: Message length.
            data: Message data
        """
        r_port = 0 if port is None else self._int(port)
        r_rssi = self._int(rssi)
        r_snr = self._int(snr)
        r_len = self._int(length)
        if r_len > 0:
            try:
                r_data = bytes.fromhex(data)
            except ValueError:
                r_data = b''
        else:
            r_data = b''
        self._downlink.append(
            {
                'port': r_port,
                'rssi': r_rssi,
                'snr': r_snr,
                'len': r_len,
                'data': r_data,
            }
        )

    def _process_events(self, timeout=None) -> None:
        """Process module event queue.

        Process event queue looking for incoming (downlink) messages.

        Args:
            timeout (optional): maximum time to wait for event.
                Defaults to None.

        Raises:
            - Rak811TimeoutError: no answer
            - Rak811ResponseError: error returned or unexpected response
                received
        """
        events = self._get_events(timeout)
        # Check for downlink
        for event in events:
            # LoRaWan format: <port>,<rssi>,<snr>,<len>[:<data>]
            # LoRa P2P format: ,<rssi>,<snr>,<len>[:<data>]
            m = match(r'((\d+),)?(-?\d+),(-?\d+),(\d+)(:(.*))?$', event)
            if m:
                _, port, rssi, snr, length, _, data = m.groups()
                self._add_downlink(port, rssi, snr, length, data)
            else:
                raise Rak811ResponseError(ErrorCode.ERR_INVALID_EVENT)

    """Generic get / set commands."""

    def set_config(self, config: str) -> List[str]:
        """Execute set_config command.

        The module will return:
            - "OK" (most cases)
            - "OK" / "<BOOT MODE>" (device:boot)
            - some info / "Initialization OK" (device:restart / lora:work_mode)

        Args:
            config: Config string to send in the format <type>:<topic>[:<param>]...
                Supported types and topics:
                - device: restart, sleep, boot, status, uart, uart_mode, gpio
                - lora: region, channel, dev_eui, app_eui, app_key, dev_addr,
                    apps_key, nwks_key, join_mode, work_mode, ch_mask, class,
                    confirm, dr, tx_power, adr, send_interval
                - lorap2p: transfer_mode, channel configuration

        Raises:
            Rak811ResponseError: Module answered with an error code.
            Rack811TimeoutError: No answer received and timeout is reached.

        Returns:
            List of responses (Informational)
        """
        # We use the "_list" variant to drain the buffer
        return self._send_command_list(f'set_config={config}')

    def get_config(self, config: str) -> List[str]:
        """Get configuration item.

        Args:
            config: Config string to send in the format <type>:<topic>[:<param>]
                Supported types and topics:
                - device: status, gpio, adc
                - lora: channel, status

        Raises:
            Rak811ResponseError: Module answered with an error code.
            Rack811TimeoutError: No answer received and timeout is reached.

        Returns:
            Module answers (list).
        """
        return(self._send_command_list(f'get_config={config}'))

    """General AT commands."""

    @property
    def version(self) -> str:
        """Get module version.

        Raises:
            - Rak811TimeoutError: no answer
            - Rak811ResponseError: error returned

        Returns:
            Module version.
        """
        return(self._send_command('version'))

    @property
    def help(self) -> List[str]:
        """Get module help.

        Raises:
            - Rak811TimeoutError: no answer
            - Rak811ResponseError: error returned

        Returns:
            Module help (list).
        """
        return(self._send_command_list('help'))

    def run(self) -> None:
        """Exit boot mode and enter normal mode.

        Raises:
            - Rak811TimeoutError: no answer
            - Rak811ResponseError: error returned
        """
        self._send_command('run')

    """Interface commands."""

    def send_uart(self, data: Union[bytes, str], index: int = 3) -> None:
        """Send data over UART.

        Args:
            data: data to be sent. If the datatype is bytes it will be send
                as such. Strings will be converted to bytes.
            index (optional): UART index to use (1, 3). Defaults to 3.
                UART1 is the AT Command interface, so you probably want to use
                UART3!

        Raises:
            - Rak811TimeoutError: no answer
            - Rak811ResponseError: error returned
        """
        if type(data) is not bytes:
            data = (bytes)(data, 'utf-8')
        data = hexlify(data).decode('ascii')

        self._send_command(f'send=uart:{index}:{data}')

    """LoRa commands"""

    def join(self) -> None:
        """Join the configured network.

        ABP requires the following parameters to be set prior join:
            - dev_addr
            - nwks_key
            - apps_key

        OTAA requires the following parameters to be set prior join:
            - dev_eui
            - app_eui
            - app_key

        Raises:
            - Rak811TimeoutError: no answer
            - Rak811ResponseError: error returned
        """
        # Extend timeout for the join command
        self._send_command('join', timeout=self._event_timeout)

    def send(self, data: Union[bytes, str], port: int = 1) -> None:
        """Send LoRaWan message.

        Args:
            data: data to be sent. If the datatype is bytes it will be send
                as such. Strings will be converted to bytes.
            port (optional): port number to use (1-223). Defaults to 1.

        Raises:
            - Rak811TimeoutError: no answer
            - Rak811ResponseError: error returned
        """
        if type(data) is not bytes:
            data = (bytes)(data, 'utf-8')
        data = hexlify(data).decode('ascii')

        self._send_command(f'send=lora:{port}:{data}', timeout=self._event_timeout)

        # Process events - Check for downlink / send confirmation
        # It is issued immediately after the "OK" response so don't have to
        # wait long.
        try:
            self._process_events(timeout=0.1)
        except Rak811TimeoutError:
            pass
        if self.nb_downlinks:
            logger.debug('Downlink available')
        else:
            logger.debug('No downlink')

    @property
    def nb_downlinks(self) -> int:
        """Get the number of downlink messages in the receive buffer.

        Returns:
            Number of messages in the downlink buffer.
        """
        return len(self._downlink)

    def get_downlink(self) -> str:
        """Get a downlink message from the receive buffer.

        Returns:
            Dictionary with the following keys:
                port: port number
                rssi: RSSI
                snr: SNR
                len: data length (0 for empty / confirmation messages)
                data: data itself (None is len is 0)
        """
        if len(self._downlink) == 0:
            return None
        else:
            return self._downlink.pop(0)

    """ LoraP2P commands."""

    def send_p2p(self, data: Union[bytes, str]) -> None:
        """Send P2P message.

        Args:
            data: data to be sent. If the datatype is bytes it will be send
                as such. Strings will be converted to bytes.

        Raises:
            - Rak811TimeoutError: no answer
            - Rak811ResponseError: error returned
        """
        if type(data) is not bytes:
            data = (bytes)(data, 'utf-8')
        data = hexlify(data).decode('ascii')

        self._send_command(f'send=lorap2p:{data}')

    def receive_p2p(self, timeout: float) -> None:
        """Wait for P2P message.

        The method wil return when one or more messages are received or when
        timeout is reached.

        The method does not return messages, use nb_downlinks and
        get_downlink() to fetch downlinks.

        Args:
            timeout: maximum time to wait.
        """
        try:
            self._process_events(timeout=timeout)
        except Rak811TimeoutError:
            pass
        if self.nb_downlinks:
            logger.debug('Message available')
        else:
            logger.debug('Nothing received')



"""RAK811 library and command line utility.

Setup file for the project

Copyright 2019 Philippe Vanhaesendonck

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

SPDX-License-Identifier: Apache-2.0
"""

from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='rak811',
    use_scm_version={"local_scheme": "no-local-version"},
    description='Interface for RAK811 LoRa module',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/AmedeeBulle/pyrak811',
    author='Philippe Vanhaesendonck',
    author_email='philippe.vanhaesendonck@e-bulles.be',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Operating System :: POSIX :: Linux',
    ],
    packages=find_packages(),
    python_requires='>=3.5',
    setup_requires=['setuptools_scm'],
    install_requires=[
        'click>=7.1',
        'pyserial',
        'RPi.GPIO; platform_machine=="armv7l" or platform_machine=="armv6l"',
        'setuptools'
    ],
    extras_require={
        'test': [
            'flake8',
            'flake8-comprehensions',
            'flake8-docstrings',
            'flake8-import-order',
            'pep8-naming',
            'pydocstyle==6.0.0',
            'pytest',
            'mock',
            'coverage',
            'tox',
        ],
    },
    entry_points={
        'console_scripts': [
            'rak811=rak811.cli:cli',
            'rak811v3=rak811.cli_v3:cli',
        ],
    },

)
